\LOGICLAB.XPL	11-Mar-2004	Lenny Boreal		loren_blaney@idcomm.com
\Digital Logic Lab, version 1.0 (also see "ShowAbout")
\Compile with XX.BAT (XPLX)

\REVISIONS:

inc	\cxpl\codesi;			\intrinsic routine declarations

def	VWidth=640, VHeight=480;	\screen dimensions (pixels)
def	Debug=false;			\true enables runtime error trapping
def	IntSize=2;			\number of bytes in an integer

\This program uses the technique of bit-plane flipping to prevent flicker. The
\ straightforward way of displaying moving objects is to erase the screen then
\ redraw the objects in their new positions. However an annoying flicker is
\ caused if the erased screen is visible even for just a sixtieth of a second.
\ This problem can be solved by dividing video memory into two parts. While one
\ part is displayed, the other is hidden from view as it is being erased and
\ redrawn. When the redrawn image is complete, the two parts are swapped, thus
\ displaying the newly drawn image. The two parts flip back and forth as the
\ objects move about the screen.
\
\Memory for video mode $12 (640x480 with 16 colors) is arranged into four
\ planes. This program divides these planes into two parts, each consisting of
\ two planes. A drawback of this is that with only two planes, the number of
\ colors that can be displayed simultaneously is limited to four. Both pairs
\ of planes are set up with identical colors, which are determined by the
\ palette and color registers.
\
\To really understand how all this works, you need to study a book such as
\ "Programmer's Guide to the EGA and VGA Cards" by Richard Ferraro.
\
\The 4 Colors:
def	Black = $0,
	Green = $5,
	Red =   $A,
	White=  $F,
	Transparent=-1;

\About 90% of the time it takes to draw a pixel is spent by merely accessing
\ the the VGA registers and memory. HUNDREDS of wait cycles are consumed. Thus
\ pixel drawing is sped up (by about a factor of four) by buffering the
\ eight pixels in a byte before they are written to the VGA. Also a copy of the
\ VGA's Bit Mask Register is maintained so that it only needs to be updated when
\ it changes. Horizontal and vertical lines don't require changing this register
\ very often.

int	DrawMask,	\selects bit planes for drawing into (see SetVGA)
	ShowMask,	\selects bit planes for displaying
	VgaBufAddr,	\offset address into Screen for buffered byte
	VgaBufColor,	\color of pixel(s) in buffered byte
	VgaBufMask,	\buffered bit mask; 1's will be set to BufColor
	VgaBitMask;	\bit mask currently in VGA register

int	PenX, PenY,	\current pen coordinates (in pixels) for DrawLine etc.
	LMargin,	\left margin: text starts at this X coordinate (pixels)
	ColorFg, ColorBg; \foreground and background colors for text sprites

def	FontWidth=6, FontHeight=12;	\font character cell dimensions (pixels)

\Object types (order must match order in ObjWidth and ObjHeight in Main):
def	NULL, AND, OR, NOT, BUT, NAND, NOR, XOR, LATCH, FLIP, NODE, CLK, TERM,
	NAND2, NOR2, NOT2, LABEL;

def	DataMax=500;	\maximum number of objects (gates etc.)

int	CpuReg,		\address of CPU register array (from GetReg)
	DataSeg,	\segment address of our data
	DataSize,	\number of objects currently active
	DataSizeB,	\backup copy for Undo command
	Dirty,		\flag: circuit has been altered and may need to be saved
	ClkSpeed,	\square-wave generator (CLK) period (1/15ths of seconds)
	ClkCtr,		\count-down timer for square wave generator clock (CLK)
	HaveGrid,	\flag: display grid pattern
	PspSeg,		\segment address of our Program Segment Prefix
	SelBox0X,	\coordinates of corners of selection box (pixels)
	SelBox0Y,
	SelBox1X,
	SelBox1Y,
	SelBox0XB,	\coordinates of corners of selection box (pixels)
	SelBox0YB,
	SelBox1XB,
	SelBox1YB;

\DoMenuCmd variables:
int	X00,		\horizontal position of previous menu or title (chars)
	TWidth0,	\width of previous menu title (chars)
	OldMouseButton;	\flag: state of mouse button on previous iteration

char	FileName($7F+4); \for file name on command line, if any
def	FileNameSize=32; \maximum number of characters in a file name, incl path

int	ObjType(DataMax),\NULL, AND, OR, NOT, BUT, etc.
	ObjX(DataMax),	 \screen coordinates
	ObjY(DataMax),
	ObjIn1(DataMax), \index of (another) object connected to object's input
	ObjIn2(DataMax), \(initially assume maximum of two inputs per gate)
	ObjOut(DataMax), \boolean: object's output
	ObjSel(DataMax), \flag: object is in selection box

	ObjTypeB(DataMax), \backup copies of object data
	ObjXB(DataMax),
	ObjYB(DataMax),
	ObjIn1B(DataMax),
	ObjIn2B(DataMax),
	ObjOutB(DataMax),
	ObjSelB(DataMax),

	ObjWidth,	 \constant array: width of object (except TERM and NODE:
			 \ this is actually the X offset to the object's output)
	ObjHeight,	 \constant array: height of object (pixels)
	Clock(DataMax);	 \initial state of clock (for clocked flip-flops)

seg char Screen(1);	 \video graphic screen (= $A000 usually)

\Mouse pointer images:
int	PtrArrow(2*16),
	PtrCross(2*16),
	PtrHand(2*16);

\Sprites:
char	SpAND(2+22*17),		\AND gate
	SpOR(2+22*17),		\OR gate
	SpNOT(2+22*17),		\inverter
	SpBUT(2+22*17),		\red button
	SpBUTOn(2+22*17),	\green button
	SpNAND(2+26*17),	\NAND gate
	SpNOR(2+26*17),		\NOR gate
	SpXOR(2+26*17),		\XOR gate
	SpLATCH(2+23*17),	\latched DE flip-flop
	SpFLIP(2+23*17),	\DC flip-flop (clocked)
	SpNODE(2+22*17),	\node (for connecting signals)
	SpCLK(2+23*17),		\square-wave generator clock
	SpTERM(2+22*17),	\red terminal (indicator)
	SpTERMOn(2+22*17),	\green terminal
	SpNAND2(2+26*17),	\DeMorganized NAND gate (bubbles on input side)
	SpNOR2(2+26*17),	\DeMorganized NOR gate
	SpNOT2(2+22*17);	\DeMorganized inverter

char	SpXUp(2+15*13),		\close button [X] in normal, up position
	SpXDn(2+15*13),		\close button in pressed position
	SpRadioOffUp(2+9*9),	\radio button: not selected, up position
	SpRadioOnUp(2+9*9),	\radio button: selected
	SpRadioOffDn(2+9*9),	\radio button: not selected, down position
	SpRadioOnDn(2+9*9),	\radio button: selected
	SpFrame1(2+12*30),	\  window frame corners and sides
	SpFrame2(2+ 6*30),	\  (numbered like keypad)
	SpFrame3(2+12*30),	\ 
	SpFrame4(2+12*12),	\ 
	SpFrame6(2+12*12),	\ 
	SpFrame7(2+12*24),	\ 
	SpFrame8(2+ 6*24),	\ 
	SpFrame9(2+30*24),	\ x
	SpFrameA(2+ 6*30),	\ [
	SpFrameB(2+ 6*30),	\ 
	SpFrameC(2+ 6*30);	\   ]

char	SpButUp(4, 2+42*14),	\Yes, No, Cancel and Print buttons
	SpButDn(4, 2+42*14);

\This must be defined last, after all the other sprites (see DrawSprite):
char	SpFont(256, 2+FontWidth*FontHeight);	\character font

def	BS=$08, Tab=$09, LF=$0A, FF=$0C, CR=$0D, Esc= $1B, Ctrl=$40;
def	Func1=$3B, Func2=$3C, Insert=$52, Delete=$53,		\scan codes
	UpArrow=$48, DnArrow=$50, LtArrow=$4B, RtArrow=$4D;

\===============================================================================

proc	SetVGA;		\Set VGA registers for bit-plane flipping
int	T;
begin
port($3C4):= 2;		\select Sequencer Map Mask Register
port($3C5):= DrawMask;	\enable bit planes for writing

T:= port($3DA);		\clear address flip-flop
port($3C0):= $32;	\select Color Plane Enable Register
			\32h = 12h + Palette Address Source (PAS) bit
port($3C0):= ShowMask;	\enable bit planes for displaying

port($3CE):= 5;		\select "write mode 2", which allows writing directly
port($3CF):= 2;		\ to the video memory (although in a very twisted way)
end;	\SetVGA



proc	DrawPix(X, Y, C);	\Draw a pixel (graphic mode $12; write mode 2)
int	X, Y, C;		\pixel coordinates and color (0..$F)
int	I, M, T;
begin
if X < 0 then return;		\clip
if X >= VWidth then return;
if Y < 0 then return;
if Y >= VHeight then return;

M:= $80 >> (X&7);			\bit mask for pixel's position in byte
I:= X>>3 + Y*80;			\byte address offset into Screen ($A000)
if I=VgaBufAddr & C=VgaBufColor then	\if byte & color are established then
	VgaBufMask:= VgaBufMask ! M	\ merely update the pending bit mask
else	begin				\actually write the buffered byte to VGA
	if VgaBufMask # VgaBitMask then	\update the Bit Mask Register only when
		begin			\ necessary
		Pout(VgaBufMask<<8 +8, $3CE, 1); \word writes are much faster
		VgaBitMask:= VgaBufMask;	 \ than two byte writes
		end;
	T:= Screen(0, VgaBufAddr);	\read to latch any existing bits in byte
	Screen(0, VgaBufAddr):= VgaBufColor; \write new bits in specified color

	VgaBufAddr:= I;			\begin new buffer for current pixel
	VgaBufColor:= C;
	VgaBufMask:= M;
	end;
end;	\DrawPix



proc	OpenPixBuf;		\Initialize pixel buffering
begin
VgaBufAddr:= -1;
VgaBufColor:= 0;
VgaBufMask:= 0;
VgaBitMask:= 0;
end;	\OpenPixBuf



proc	ClosePixBuf;		\Draw any pixels currently being buffered
int	T;
begin
Pout(VgaBufMask<<8 +8, $3CE, 1);	\set Bit Mask Register
T:= Screen(0, VgaBufAddr);		\read to latch any existing bits in byte
Screen(0, VgaBufAddr):= VgaBufColor;	\write new bits with specified color
OpenPixBuf;
end;	\ClosePixBuf



proc	DrawLine(X1, Y1, C);	\Draw line from current pen position to X1,Y1
\Warning: This only draws either horizontal or vertical lines
int	X1, Y1,
	C;	\Color (in low byte); the high byte specifies dash pattern
int	X0, Y0, X, Y, T;
int	Dash, Mask;
begin
X0:= PenX;
PenX:= X1;
if X1 < X0 then [T:= X0;   X0:= X1;   X1:= T];

Y0:= PenY;
PenY:= Y1;
if Y1 < Y0 then [T:= Y0;   Y0:= Y1;   Y1:= T];

Dash:= C >> 8;
C:= C & $00FF;
Mask:= $80;

for X:= X0, X1 do
	begin
	if Mask & Dash then [] else DrawPix(X, Y0, C);
	Mask:= Mask >> 1;
	if Mask = 0 then Mask:= $80;
	end;

for Y:= Y0, Y1 do
	begin
	if Mask & Dash then [] else DrawPix(X0, Y, C);
	Mask:= Mask >> 1;
	if Mask = 0 then Mask:= $80;
	end;
end;	\DrawLine



proc	MovePen(X, Y);		\Set pen position
int	X, Y;	\coordinates (pixels)
begin
PenX:= X;
PenY:= Y;
end;	\MovePen



func	ChkPix(X, Y, C);	\Returns 'true' if pixel at X,Y matches color C
int	X, Y, C;		\(much faster than ReadPix, which calls BIOS)
int	B;
begin
Pout(C<<8+2, $3CE, 1);		\set Color Compare Register
Pout($A<<8+5, $3CE, 1);		\enable read mode 1 while keeping write mode 2
Pout(ShowMask<<8+7, $3CE, 1);	\set Color Do(n't) Care Reg with visible planes
B:= Screen(0, Y*80 + X>>3);	\get byte at coordinates X,Y
return (B & $80>>(X&7)) # 0;	\test pixel at X
end;	\ChkPix

\===============================================================================

proc	Exit(Msg);		\Display (error) message and terminate program
char	Msg;
begin
SetVid($03);			\restore normal text mode
Text(0, Msg);
Crlf(0);
exit;
end;	\Exit



proc	Beep;			\A not-too-obnoxious beep
begin
Sound(false, 1, 1000);		\synchronize with system timer to make tone a
Sound(true, 1, 3000);		\ consistent duration and a consistent sound
end;	\Beep



func	CallInt(Int, AX, BX, CX, DX, BP, DS, ES); \Call a software interrupt
int	Int, AX, BX, CX, DX, BP, DS, ES; \(unused arguments need not be passed)
begin
CpuReg(0):= AX;
CpuReg(1):= BX;
CpuReg(2):= CX;
CpuReg(3):= DX;
CpuReg(6):= BP;
CpuReg(9):= DS;
CpuReg(11):= ES;
Softint(Int);
return CpuReg(0);		\return AX register
end;	\CallInt



func	GetKey;			\Get character from keyboard (wait if necessary)
int	SC, Ch;			\this is a low-level routine with no echo, no
begin				\ Ctrl-C, and no flashing cursor.
SC:= CallInt($16, $0000);	\read character and scan code from keyboard
Ch:= SC & $00FF;
if Ch = 0 then Ch:= -(SC>>8);	\return non-ASCII chars as negative scan code
return Ch;
end;	\GetKey



proc	InsertKey(Ch);		\Insert character into the keyboard buffer
int	Ch;
CallInt($16, $0500, 0, Ch);



proc	Delay(N);		\Delay N sixtieths of a second
int	N;			\(negative N does 0 delay)
int	I;
begin
for I:= 1, N do
	begin
	while port($3DA) & $08 do;		\wait for no vertical retrace
	repeat until port($3DA) & $08;		\wait for vertical retrace
	end;
end;	\Delay



func	ToLower(Ch);		\Convert character to lowercase
int	Ch;
return if Ch>=^A & Ch<=^Z then Ch!$20 else Ch;



func	StrNCmp(A, B, N);	\Compare string A to string B up to N bytes long
\This returns:
\	>0 if A > B
\	=0 if A = B
\	<0 if A < B
\This provides a general string compare, for example:
\ if StrNCmp(A, B, N) >= 0 then...	(if A >= B then...)
char	A, B;	\Strings to be compared (must be right justified)
int	N;	\Number of bytes to compare
int	I;
begin
for I:= 0, N-1 do
	if A(I) # B(I) then
		return A(I) - B(I);
return 0;			\They're equal
end;	\StrNCmp

\=============================== MOUSE ROUTINES ================================

func	OpenMouse;		\Initializes mouse; returns 'false' if it fails
begin				\pointer is set to center of screen but is hidden
CallInt($21, $3533);		\make sure mouse vector ($33) points to something
if CpuReg(1)=0 & CpuReg(11)=0 then return false;
return CallInt($33, $0000);	\reset mouse; return 'false' if failure
end;	\OpenMouse



proc	ShowMouse(On);		\Turn mouse pointer on or off
\The video mode should be set before calling this routine.
\This counts the number of times the pointer is turned on or off. If the
\ pointer is turned on twice, it must be turned off twice before it
\ actually goes off. The pointer should be turned off before drawing over
\ it and before your program exits. Setting the video mode will also turn
\ off the pointer.
int	On;	\Flag: True = pointer on; False = pointer off
CallInt($33, if On then $0001 else $0002);



func	GetMousePosition(N); \Return position of specified mouse coordinate
int	N;	\0 = X coordinate; 1 = Y coordinate
\For video modes $0-$E and $13 the maximum coordinates are 639x199, minus
\ the size of the pointer. For modes $F-$12 the coordinates are the same as
\ the pixels. For 80-column text modes divide the mouse coordinates by 8 to
\ get the character cursor position.
begin
CallInt($33, $0003);
return if N then CpuReg(3) else CpuReg(2);
end;	\GetMousePosition



func	GetMouseButton;	  \Returns 'true' if a mouse button or shift key is down
int	Sh;
begin
Sh:= CallInt($16, $0200) & $03;
CallInt($33, $0003);
return (CpuReg(1) & $03 ! Sh) # 0;
end;	\GetMouseButton

\===============================================================================

proc	Hilite(X0, Y0, W, H, C); \Highlight non-black area in specified color
int	X0, Y0, W, H, C;	 \coordinates, width & height (pixels), color
int	X, Y;
begin
ShowMouse(false);
for Y:= Y0, Y0+H-1 do
    for X:= X0, X0+W-1 do
	if ~ChkPix(X, Y, Black) then DrawPix(X, Y, C);
ClosePixBuf;
ShowMouse(true);
end;	\Hilite



proc	DrawSelBox;		\Draw selection box
begin
ShowMouse(false);
MovePen(SelBox0X, SelBox0Y);
DrawLine(SelBox1X, SelBox0Y, $0F00+White);
DrawLine(SelBox1X, SelBox1Y, $0F00+White);
DrawLine(SelBox0X, SelBox1Y, $0F00+White);
DrawLine(SelBox0X, SelBox0Y, $0F00+White);
ClosePixBuf;
ShowMouse(true);
end;	\DrawSelBox



proc	DrawGrid;		\Fill screen with grid pattern
int	X, Y, I, T;
begin
ShowMouse(false);
port($3CE):= 8;				\select Bit Mask Register
port($3CF):= $80;			\set every 4th bit
I:= 0;
for Y:= 0, VHeight/8-1 do		\for every 4th line...
    begin
    for X:= 0, VWidth>>3 -1 do
	begin
	T:= Screen(0, I);		\read to latch any existing bits in byte
	Screen(0, I):= if (X&3)=0 & (Y&3)=0 then White else Red;
	I:= I + 1;
	end;
    I:= I + 7*(Vwidth>>3);
    end;
ShowMouse(true);
end;	\DrawGrid



proc	EraseScreen;		\Clear the video screen (in write mode 2)
int	X, Y, I;
begin
ShowMouse(false);
DrawMask:= ~DrawMask & $F;		\start drawing alternate pair of planes
SetVGA;
port($3CE):= 8;				\select Bit Mask Register
port($3CF):= $FF;			\set all bits to erase entire byte
I:= 0;
for Y:= 0, VHeight-1 do			\for all displayed video memory...
    for X:= 0, VWidth>>3 -1 do
	begin
	Screen(0, I):= Black;		\write byte with black
	I:= I + 1;
	end;
OpenPixBuf;
ShowMouse(true);
end;	\EraseScreen

\-------------------------------------------------------------------------------

proc	DrawSprite(X0, Y0, S, Trans);	\Draw a sprite
int	X0, Y0;	\coordinates of upper-left corner
char	S;	\address of sprite data
int	Trans;	\flag: transparent background (versus opaque)
int	IsCh;	\flag: sprite is a character
int	X, Y, K, C, W, H;
begin
IsCh:= S - addr SpFont(0,0) >= 0;	\assumes SpFont is defined last
\(Don't be tempted to optimize this comparison: addresses are UNsigned values.)

W:= S(0);	\get width and height (in pixels)
H:= S(1);
K:= 2;
for Y:= Y0, Y0+H-1 do
    for X:= X0, X0+W-1 do
	begin
	C:= S(K);			\get pixel's color
	K:= K + 1;
	if C ! ~Trans then		\background (0) is transparent
	  begin
	  if IsCh then
	    C:= if C then ColorFg else ColorBg;
	  DrawPix(X, Y, C);
	  end;
	end;
end;	\DrawSprite



proc	DrawObj(X, Y, Obj);	\Draw the sprite that cooresponds to an object
int	X, Y, Obj;
char	Sp;
begin
case Obj of
  AND:	Sp:= SpAND;
  OR:	Sp:= SpOR;
  NOT:	Sp:= SpNOT;
  BUT:	Sp:= SpBUT;
  NAND:	Sp:= SpNAND;
  NOR:	Sp:= SpNOR;
  XOR:	Sp:= SpXOR;
  LATCH:Sp:= SpLATCH;
  FLIP:	Sp:= SpFLIP;
  NODE:	Sp:= SpNODE;
  CLK:	Sp:= SpCLK;
  TERM:	Sp:= SpTERM;
  NAND2:Sp:= SpNAND2;
  NOR2:	Sp:= SpNOR2;
  NOT2:	Sp:= SpNOT2
other	[];

DrawSprite(X, Y, Sp, true);
ClosePixBuf;
end;	\DrawObj



proc	DrawChar(Ch);		\Draw an ASCII character at current pen position
int	Ch;
begin
case Ch of
  CR:	begin
	PenX:= LMargin;
	end;
  LF:	PenY:= PenY + FontHeight
other	begin
	DrawSprite(PenX, PenY, addr SpFont(Ch,0), ColorBg=Transparent);
	PenX:= PenX + FontWidth;
	end;
end;	\DrawChar



proc	DrawText(Str);		\Display text string at current pen position
char	Str;
int	I, Ch;
begin
I:= 0;
loop	begin
	Ch:= Str(I);
	I:= I + 1;
	if Ch >= $80 then quit;
	DrawChar(Ch);
	end;
DrawChar(Ch & $7F);
end;	\DrawText



proc	DrawRawText(Str);	\Print string containing an extended character
char	Str;
int	I, Ch;
begin
I:= 0;
loop	begin
	Ch:= Str(I);
	I:= I + 1;
	if Ch = $A0 then quit;
	DrawChar(Ch);
	end;
end;	\DrawRawText



proc	DrawNum(N);	\Draw (display) positive integer at current pen location
int	N;
int	R;
begin
N:= N / 10;
R:= Rem(0);
if N then DrawNum(N);
DrawChar(R+^0);
end;	\DrawNum



proc	DrawTitleText(X0, Y0, Str);	\Display a title in bold with shadow
\Warning: This alters the colors, and it does not set the pen position to the
\ end of the text like DrawText does.
int	X0, Y0;		\coordinates of upper left corner of first character
char	Str;		\title string
int	I, J, Ch;
begin
ColorFg:= Black;   ColorBg:= Transparent;
for J:= 0, 3 do		\W's & M's look bad with 3; everything looks bad with 2
	begin
	MovePen(X0+2, Y0+1);
	PenX:= PenX - (J+1)/2;		\PenX - 0, 1, 1, 2
	if J >= 2 then
		begin
		PenY:= PenY - 1;
		ColorFg:= White;   ColorBg:= Transparent;
		end;
	I:= 0;
	loop	begin
		Ch:= Str(I);
		I:= I + 1;
		if Ch >= $80 then quit;
		DrawChar(Ch);
		PenX:= PenX + 1;	\extra pixel between characters
		end;
	DrawChar(Ch & $7F);
	end;
end;	\DrawTitleText

\-------------------------------------------------------------------------------

proc	DrawMenuBar;		\Draw menu bar across top of screen
int	I, L, Ch;
char	Str(FileNameSize);
begin
ShowMouse(false);
MovePen(0, 0);
ColorFg:= Black;   ColorBg:= White;
DrawText("  File  Edit  View  Parts  Help          ");
ColorBg:= Green;
DrawText("|                      ");
X00:= -1;		\indicate highlight (if any) is gone

\Display current input file name, if there is one, otherwise display "LogicLab"
L:= 0;			\(for safety)
for I:= 0, 18 do	\there is room to display a path name of up to 18 chars
	begin
	Ch:= FileName(I);
	if Ch = ^. then			\save name length, so it can be centered
		[L:= I;  I:= FileNameSize];
	Str(I):= Ch;
	end;
if L > 0 then
	begin
	Str(L-1):= Str(L-1) ! $80;
	DrawTitleText((51-L/2)*FontWidth, 0, Str);
	end
else	DrawTitleText((51-8/2)*FontWidth, 0, "LogicLab");

PenX:= 63*FontWidth;
ColorFg:= Black;   ColorBg:= White;
DrawText("|                                         ");

MovePen(75*FontWidth, 0);		\also see DnD
DrawNum(GetMousePosition(0)/4);   DrawChar(^,);
DrawNum(GetMousePosition(1)/4);
MovePen(90*FontWidth, 0);
DrawNum(DataMax-DataSize);

MovePen(0, FontHeight);
DrawLine(VWidth-1-4-3+1, FontHeight, Green);		\shadow
DrawPix(41*FontWidth, FontHeight, Black);
DrawPix(63*FontWidth, FontHeight, Black);

DrawSprite(VWidth-15, 0, SpXUp, false);			\ [X] button
ClosePixBuf;
ShowMouse(true);
end;	\DrawMenuBar



proc	DrawFrame(X0, Y0, W, H, Title);	\Draw a window frame
int	X0, Y0,		\upper-left corner of text in window (char cells)
	W, H;		\width and height of text (char cells)
char	Title;		\string: title of window frame
int	XL,		\coordinate of left edge of of window frame (pixels)
	XR,		\coordinate of right edge
	YT,		\coordinate of top edge
	YB,		\coordinate of top edge of bottom of frame
	I, J;
begin
XL:= X0*FontWidth-12;	\position of left edge of window frame (pixels)
XR:= (X0+W)*FontWidth;	\position of right edge of window frame (pixels)
YT:= Y0*FontHeight-24;	\Y coord of top edge of window frame (pixels)

for J:= 0, H*FontHeight-1 do		\make white background
	begin				\horizontal lines are fast
	MovePen(X0*FontWidth, Y0*FontHeight+J);
	DrawLine(XR-1, Y0*FontHeight+J, White);
	end;

DrawSprite(XL, YT, SpFrame7, false);
for I:= 0, W-1-3 do
	DrawSprite(XL+12+I*FontWidth, YT, SpFrame8, false);
DrawSprite(XR-18, YT, SpFrame9, false);

for I:= 0, H-1 do
	begin				\draw sides
	DrawSprite(XL, YT+24+I*FontHeight, SpFrame4, false);
	DrawSprite(XR, YT+24+I*FontHeight, SpFrame6, false);
	end;

YB:= (Y0+H)*FontHeight;
DrawSprite(XL, YB, SpFrame1, false);
for I:= 0, W-1 do
	DrawSprite(XL+12+I*FontWidth, YB, SpFrame2, false);
DrawSprite(XR, YB, SpFrame3, false);

ColorFg:= Black;   ColorBg:= Green;
MovePen((X0+2)*FontWidth, (Y0-1)*FontHeight-6);
DrawTitleText(PenX, PenY, Title);
ClosePixBuf;
end;	\DrawFrame

\===============================================================================

proc	DrawConn(InX, InY, OutX, OutY, Lev);	\Draw a connecting signal line
int	InX, InY,	\coordinates of gate's input (pixels)
	OutX, OutY,	\coordinates of gate's output (or end being dragged)
	Lev;		\boolean: level (high/low)
int	C, HX, HY;
begin
C:= if Lev then Green else Red;
MovePen(OutX, OutY);
if OutX <= InX then			\ Out --------
	begin				\	      |
	HX:= (InX+OutX)/2;		\	      |
	DrawLine(HX, OutY, C);		\	       ----------- In
	DrawLine(HX, InY, C);
	DrawLine(InX, InY, C);
	end
else	begin				\		        Out -
	HY:= (InY+OutY)/2;		\			     |
	DrawLine(OutX+1, OutY, C);	\			     |
	DrawLine(OutX+1, HY, C);	\     -----------------------
	DrawLine(InX-1, HY, C);		\    |
	DrawLine(InX-1, InY, C);	\    |
	DrawLine(InX, InY, C);		\     - In
	end;
ClosePixBuf;
end;	\DrawConn



proc	DrawCircuit;		\Display the complete circuit
int	I, J, W;
begin
ShowMouse(false);
EraseScreen;
if HaveGrid then DrawGrid;

for I:= 0, DataSize-1 do	\draw the objects...
	begin
	if ObjType(I)=BUT & ObjOut(I) then
		begin
		DrawSprite(ObjX(I), ObjY(I), SpBUTOn, true);
		end
	else if ObjType(I)=TERM & ObjOut(I) then
		begin
		DrawSprite(ObjX(I), ObjY(I), SpTERMOn, true);
		end
	else if ObjType(I)=LABEL then
		begin
		MovePen(ObjX(I), ObjY(I));
		ColorFg:= White;   ColorBg:= Transparent;
		DrawChar(ObjOut(I) & $00FF);
		DrawChar(ObjOut(I) >> 8);
		DrawChar(ObjIn1(I) & $00FF);
		DrawChar(ObjIn1(I) >> 8);
		DrawChar(ObjIn2(I) & $00FF);
		DrawChar(ObjIn2(I) >> 8);
		end
	else	DrawObj(ObjX(I), ObjY(I), ObjType(I));
	end;

for I:= 0, DataSize-1 do	\draw the connections on top of the objects...
	begin
	if ObjType(I)=AND ! ObjType(I)=OR ! ObjType(I)=NAND ! ObjType(I)=NOR !
	   ObjType(I)=XOR ! ObjType(I)=LATCH ! ObjType(I)=FLIP !
	   ObjType(I)=NAND2 ! ObjType(I)=NOR2 then
		begin
		J:= ObjIn1(I);	\index of object that is connected to input 1
		if J >= 0 then
		  begin
		  W:= ObjWidth(ObjType(J));
		  DrawConn(ObjX(I), ObjY(I)+4, ObjX(J)+W, ObjY(J)+8, ObjOut(J));
		  \     input1's coordinates, coords of what it's connected to
		  end;
		J:= ObjIn2(I);	\index of object that is connected to input 2
		if J >= 0 then
		  begin
		  W:= ObjWidth(ObjType(J));
		  DrawConn(ObjX(I), ObjY(I)+12, ObjX(J)+W, ObjY(J)+8, ObjOut(J));
		  \     input2's coordinates, coords of what it's connected to
		  end;
		end
	else if ObjType(I)=NOT ! ObjType(I)=NOT2 then
		begin
		J:= ObjIn1(I);	\index of object that is connected to input 1
		if J >= 0 then
		  begin
		  W:= ObjWidth(ObjType(J));
		  DrawConn(ObjX(I), ObjY(I)+8, ObjX(J)+W, ObjY(J)+8, ObjOut(J));
		  \     input1's coordinates, coords of what it's connected to
		  end;
		end
	else if ObjType(I)=NODE ! ObjType(I)=TERM then
		begin
		J:= ObjIn1(I);	\index of object that is connected to input 1
		if J >= 0 then
		  begin
		  W:= ObjWidth(ObjType(J));
		  DrawConn(ObjX(I)+4, ObjY(I)+8, ObjX(J)+W, ObjY(J)+8, ObjOut(J));
		  \     input1's coordinates, coords of what it's connected to
		  end;
		end;
	end;

if SelBox1Y ! SelBox0Y then	\box exists (it might be expanding upside down)
    if SelBox1X#SelBox0X ! SelBox1Y#SelBox0Y then	\not a single point
	DrawSelBox;

DrawMenuBar;
ClosePixBuf;
ShowMask:= ~ShowMask & $F;
SetVGA;
ShowMouse(true);
end;	\DrawCircuit

\-------------------------------------------------------------------------------

proc	CalcCircuit;		\Calculate states of all outputs
int	I, J, K, I1, I2, Trig;
int	TenOut(DataMax);	\tenative output for clocked flip-flops



proc	Propagate(Cnt);		\Determine output states of all gates (ObjOut)
int	Cnt;
int	K, I, J, I1, I2;
begin
\Propagate changes through all combinational logic
for K:= 0, Cnt-1 do
    for I:= 0, DataSize-1 do	\for all objects...
	begin
	J:= ObjIn1(I);		\index of object that is connected to input 1
	I1:= if J >= 0 then ObjOut(J) else true;  \unconnected inputs float high
	J:= ObjIn2(I);
	I2:= if J >= 0 then ObjOut(J) else true;

	case ObjType(I) of
	  AND:	ObjOut(I):= I1 & I2;
	  OR:	ObjOut(I):= I1 ! I2;
	  NOT, NOT2:
		ObjOut(I):= ~I1;
	  BUT:	[];			\ObjOut(I):= ObjOut(I)
	  NAND, NAND2:
		ObjOut(I):= ~(I1 & I2);
	  NOR, NOR2:
		ObjOut(I):= ~(I1 ! I2);
	  XOR:	ObjOut(I):= I1 | I2;
	  LATCH:if I2 then ObjOut(I):= I1;
	  FLIP:	[];
	  NODE:	ObjOut(I):= I1;
	  CLK:	[];			\ObjOut(I):= ObjOut(I)
	  TERM:	ObjOut(I):= I1;		\acts same as a NODE; output = color
	  LABEL:[]
	other	ObjOut(I):= true;	\NULL (or undefined) object outputs high
	end;
end;	\Propagate



begin	\CalcCircuit
for K:= 0, DataSize-1 do	\worst case for propagating all ripple clocks
	begin
	Propagate((DataSize+1)&$FFFE);	\do an even number of times (see below)

	\Determine outputs of all clocked flip-flops (FLIP)
	Trig:= false;
	for I:= 0, DataSize-1 do
	    if ObjType(I) = FLIP then
		begin
		J:= ObjIn1(I);	\index of object that is connected to D input
		I1:= if J >= 0 then ObjOut(J) else true;
		J:= ObjIn2(I);			\unconnected inputs float high
		I2:= if J >= 0 then ObjOut(J) else true;  \current clock state
		TenOut(I):= ObjOut(I);		\tenative output= current output
		if J >= 0 then			\ unless positive clock edge
		    if I2 & ~Clock(J) then	\positive transition occurred
			[Trig:= true;   TenOut(I):= I1];
		end;

	\Save initial state of all possible clock inputs
	for I:= 0, DataMax-1 do Clock(I):= ObjOut(I);

	\Set all FLIP outputs at the same time
	for I:= 0, DataSize-1 do
	    if ObjType(I) = FLIP then
		ObjOut(I):= TenOut(I);

	if ~Trig then K:= DataSize;	\exit 'for' loop if no change occurred
	end;				\ (this is merely to speed things up)

\Iterate combinational logic a total of an odd number of times so feedback
\ oscillators will oscillate (such as a NOT gate feeding back on itself).
\ The above code iterated an even number of times (since even*even=even and
\ even*odd=even), so iterate one more time.
Propagate(1);
end;	\CalcCircuit

\================================== COMMANDS ===================================

proc	DelNULLs;		\Delete NULLs from data
int	I;
int	J;
begin
I:= 0;
while I < DataSize do
	begin
	if ObjType(I) = NULL then
		begin
		\Move everything above I down one place
		for J:= I, DataSize-2 do
			begin
			ObjType(J):= ObjType(J+1);
			ObjX(J):= ObjX(J+1);
			ObjY(J):= ObjY(J+1);
			ObjIn1(J):= ObjIn1(J+1);
			ObjIn2(J):= ObjIn2(J+1);
			ObjOut(J):= ObjOut(J+1);
			ObjSel(J):= ObjSel(J+1);
			end;
		DataSize:= DataSize - 1;
		\Decrement all references to inputs above I
		for J:= 0, DataSize-1 do
		    if ObjType(J) # LABEL then
			begin
			if ObjIn1(J) > I then ObjIn1(J):= ObjIn1(J) - 1;
			if ObjIn2(J) > I then ObjIn2(J):= ObjIn2(J) - 1;
			end;
		end
	else	I:= I + 1;	\(object moved down might be a NULL)
	end;
for I:= 0, DataSize-1 do Clock(I):= ObjOut(I);
end;	\DelNULLs



proc	DelObj(I);		\Delete object I
int	I;
int	J;
begin
ObjType(I):= NULL;				\make it a NULL object
for J:= 0, DataSize-1 do			\for all objects...
	begin					\if object's input is the
	if ObjIn1(J) = I then ObjIn1(J):= -1;	\ deleted object then delete
	if ObjIn2(J) = I then ObjIn2(J):= -1;	\ its input connection
	end;
end;	\DelObj



proc	DelSelBox;		\Delete selection box
int	I;
begin
SelBox0X:= 0;
SelBox0Y:= 0;
SelBox1X:= 0;
SelBox1Y:= 0;
for I:= 0, DataSize-1 do	\deselect everything
	ObjSel(I):= false;
end;	\DelSelBox



proc	FlipClks;		\Flip state of all clock generators
int	I, F;
begin
\Save initial state of all possible clock inputs
for I:= 0, DataMax-1 do Clock(I):= ObjOut(I);

F:= false;
for I:= 0, DataSize-1 do
    if ObjType(I) = CLK then
	begin
	ObjOut(I):= ~ObjOut(I);	\flip state
	F:= true;
	end;
if F then			\something changed so show result
	[CalcCircuit;   DrawCircuit];
end;	\FlipClks



proc	InsertData(FN);		\Insert object data from specified file
char	FN;			\ into the schematic at the mouse position
int	Hand, I, N, Size,
	MX, MY;
begin
Trap(false);
Hand:= Fopen(FN, 0);		\get input handle
if Geterr \#0\ then Beep	\if file not found, don't throw a fit
else	begin
	DelSelBox;
	Fset(Hand, ^I);
	OpenI(3);
	Size:= Intin(3);	\number of objects to insert
	if DataSize+Size > DataMax then
		Beep		\(don't load partial data)
	else	begin
		MX:= GetMousePosition(0) & ~3;		\4x4 grid
		MY:= GetMousePosition(1) & ~3;

		for I:= DataSize, DataSize+Size-1 do
			begin
			ObjType(I):= Intin(3);
			ObjX(I):= Intin(3) + MX;
			ObjY(I):= Intin(3) + MY;
			N:= Intin(3);
			ObjIn1(I):= if N>=0 & ObjType(I)#LABEL then N+DataSize
				else N;
			N:= Intin(3);
			ObjIn2(I):= if N>=0 & ObjType(I)#LABEL then N+DataSize
				else N;
			ObjOut(I):= Intin(3);
			Clock(I):= ObjOut(I);
			ObjSel(I):= true;
			end;
		DataSize:= DataSize + Size;
		Dirty:= true;
		end;
	FClose(Hand);
	end;
Trap(Debug);
end;	\InsertData



proc	LoadData(FN);		\Load object data from specified file
char	FN;
int	Hand, I;
begin
Trap(false);
Hand:= Fopen(FN, 0);		\get input handle
if Geterr \#0\ then Beep	\if file not found, don't throw a fit
else	begin
	for I:= 0, FileNameSize-1 do		\get name to display on menu bar
		FileName(I):= FN(I);

	DelSelBox;
	Fset(Hand, ^I);
	OpenI(3);

	DataSize:= Intin(3);
	for I:= 0, DataSize-1 do
		begin
		ObjType(I):= Intin(3);
		ObjX(I):= Intin(3);
		ObjY(I):= Intin(3);
		ObjIn1(I):= Intin(3);
		ObjIn2(I):= Intin(3);
		ObjOut(I):= Intin(3);
		Clock(I):= ObjOut(I);
		ObjSel(I):= false;
		end;

	FClose(Hand);
	Dirty:= false;
	end;
Trap(Debug);
end;	\LoadData



proc	SaveData(FN);		\Save object data into specified file name
char	FN;
int	Hand, I;
begin
Trap(false);
Hand:= Fopen(FN, 1);		\get output handle
if Geterr \#0\ then Beep	\if illegal, don't lose the circuit
else	begin
	Fset(Hand, ^O);
	Openo(3);

	Intout(3, DataSize); Crlf(3);
	for I:= 0, DataSize-1 do
		begin
		Intout(3, ObjType(I)); Chout(3, Tab);
		Intout(3, ObjX(I)); Chout(3, Tab);
		Intout(3, ObjY(I)); Chout(3, Tab);
		Intout(3, ObjIn1(I)); Chout(3, Tab);
		Intout(3, ObjIn2(I)); Chout(3, Tab);
		Intout(3, ObjOut(I)); Crlf(3);
		end;

	Close(3);
	FClose(Hand);
	Dirty:= false;
	end;
Trap(Debug);
end;	\SaveData



proc	Undo;		\Swap object data with backup arrays
int	I, T;
begin
T:= DataSizeB;	DataSizeB:= DataSize;	DataSize:= T;
for I:= 0, DataMax-1 do
	begin
	T:= ObjTypeB(I);	ObjTypeB(I):= ObjType(I);	ObjType(I):= T;
	T:= ObjXB(I);		ObjXB(I):= ObjX(I);		ObjX(I):= T;
	T:= ObjYB(I);		ObjYB(I):= ObjY(I);		ObjY(I):= T;
	T:= ObjIn1B(I);		ObjIn1B(I):= ObjIn1(I);		ObjIn1(I):= T;
	T:= ObjIn2B(I);		ObjIn2B(I):= ObjIn2(I);		ObjIn2(I):= T;
	T:= ObjOutB(I);		ObjOutB(I):= ObjOut(I);		ObjOut(I):= T;
	T:= ObjSelB(I);		ObjSelB(I):= ObjSel(I);		ObjSel(I):= T;
	end;
T:= SelBox0XB;		SelBox0XB:= SelBox0X;		SelBox0X:= T;
T:= SelBox0YB;		SelBox0YB:= SelBox0Y;		SelBox0Y:= T;
T:= SelBox1XB;		SelBox1XB:= SelBox1X;		SelBox1X:= T;
T:= SelBox1YB;		SelBox1YB:= SelBox1Y;		SelBox1Y:= T;
CalcCircuit;
DrawCircuit;
end;	\Undo



proc	Backup;		\Save object data into backup arrays
int	I;
begin
DataSizeB:= DataSize;
for I:= 0, DataSize-1 do
	begin
	ObjTypeB(I):= ObjType(I);
	ObjXB(I):= ObjX(I);
	ObjYB(I):= ObjY(I);
	ObjIn1B(I):= ObjIn1(I);
	ObjIn2B(I):= ObjIn2(I);
	ObjOutB(I):= ObjOut(I);
	ObjSelB(I):= ObjSel(I);
	end;
SelBox0XB:= SelBox0X;
SelBox0YB:= SelBox0Y;
SelBox1XB:= SelBox1X;
SelBox1YB:= SelBox1Y;
end;	\Backup

\===============================================================================

proc	ShowAbout;	\Display "About" information
def	FW=53, FH=21,			\frame width and height (char cells)
	LM=(VWidth/FontWidth-FW)/2,	\text left margin (char cells)
	TM=(VHeight/FontHeight-FH)/2;	\text top margin (char cells)
begin
ShowMouse(false);
DrawFrame(LM, TM, FW, FH, "About");
ColorFg:= Black;   ColorBg:= Transparent;
LMargin:= LM*FontWidth;
MovePen(LMargin, TM*FontHeight);

DrawText("
 LogicLab version 1.0

 Copyleft (c) 2004 Lenny Boreal

 This program is free software. You may redistribute
 it and modify it under the terms of the GNU General
 Public License version 2 as published by the Free
 Software Foundation.

 This program is distributed in the hope that it
 will be useful, but WITHOUT ANY WARRANTY; without
 even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 General Public License (LICENSE.TXT) for details.

 Beta testing by the 6502 Group (www.6502group.org).

 Lenny Boreal can be reached c/o Loren Blaney at:
 loren_blaney@idcomm.com

");
ClosePixBuf;

CallInt($33, 9, 0, 0, PtrArrow, 0, 0, DataSeg);		\set mouse pointer image
ShowMouse(true);
loop	begin
	if ChkKey then [OpenI(0);   quit];
	if GetMouseButton then
	    begin
	    ShowMouse(false);
	    DrawSprite((LM+FW)*FontWidth-14, TM*FontHeight-18, SpXDn, false);
	    ShowMouse(true);
	    while GetMouseButton do [];
	    quit;
	    end;
	end;
end;	\ShowAbout

\-------------------------------------------------------------------------------

proc	ShowHelp;	\Display "Help" information
def	FW=96, FH=35,			\frame width and height (char cells)
	LM=(VWidth/FontWidth-FW)/2,	\text left margin (char cells)
	TM=(VHeight/FontHeight-FH)/2;	\text top margin (char cells)
begin
ShowMouse(false);
DrawFrame(LM, TM, FW, FH, "Help");
ColorFg:= Black;   ColorBg:= Transparent;
LMargin:= LM*FontWidth;
MovePen(LMargin, TM*FontHeight);

DrawText(" LogicLab is used to simulate simple logic
 circuits. It enables you to quickly connect
 gates and flip-flops to see how they operate.

 The logical state of connecting lines is
 indicated by color. Green indicates a 'true'
 (or 1, or high voltage), while red indicates
 'false' (or 0, or low voltage).

 All of the menu items have shortcut keys. For
 instance, ^"a^" is used to insert an AND gate
 at the mouse pointer.

 Gates (or any objects) can be dragged into
 position and connected together. Connections
 are made by dragging an input to another
 object. Dragging an output to an input does
 not form a connection. Sometimes connections
 don't route where you want them, so nodes
 (E key) can be used to reroute them.

 Circuits can be activated either with push
 buttons (B) or with clock generators (C). The
 state of a button (B) is flipped by clicking
 the mouse on it. Keys 0 through 9 set the
 frequency of the clock generators.

 The most recently added object can be deleted
 with the Delete key (or menu item). Other ob-
 jects can be deleted by first selecting them.

 Objects are selected by pressing the mouse
 button on an empty region of the screen and,
 while holding the button down, moving the
 pointer across them. A box will appear.");

LMargin:= (LM+49)*FontWidth;
MovePen(LMargin, TM*FontHeight);

DrawText("The selection box is also used with the
Duplicate and DeMorgan functions. DeMorgan
(named after the English logician) converts
NAND, NOR and NOT gates to their equivalents.

The arrow keys can be used to make fine
adjustments on the position of the mouse
pointer. The Shift keys work the same as the
mouse buttons. Windows' cut and paste keys
(Ctrl+X and Ctrl+V) are an another way to
Delete and Duplicate.

When LogicLab exits, your circuit is auto-
matically saved in a file called _backup_.lab.

Windows users might be disappointed with the
appearance of the display. 60Hz flicker can be
reduced by turning the brightness down. An old
style CRT monitor usually looks better than a
flat panel LCD. Full-screen mode (Alt+Enter)
should be used when running under Windows 98,
otherwise it's ridiculously slow.

Printing problems with Windows can be solved
by installing a Generic / Text Only driver.

LogicLab is intended for simple circuits. It
only has 2-input gates. It does not understand
gate propagation delay, so for instance a
master-slave flip-flop cannot be made from
discrete gates. However a clocked flip-flop is
provided (F key), and it works in both syn-
chronous and ripple-clock circuits. If you
want a sophisticated (and free) logic simu-
lator, one can be downloaded from Xilinx.com.");

ClosePixBuf;

CallInt($33, 9, 0, 0, PtrArrow, 0, 0, DataSeg);		\set mouse pointer image
ShowMouse(true);
loop	begin
	if ChkKey then [OpenI(0);   quit];
	if GetMouseButton then
	    begin
	    ShowMouse(false);
	    DrawSprite((LM+FW)*FontWidth-14, TM*FontHeight-18, SpXDn, false);
	    ShowMouse(true);
	    while GetMouseButton do [];
	    quit;
	    end;
	end;
end;	\ShowHelp

\===============================================================================

proc	DoButtons(Num, ButCh, ButX, ButY, ButW, ButH, SpUp, SpDn);
\Handle soft buttons: wait for one to be clicked on, or its equivalent keystroke
int	Num,		\number of buttons
	ButCh,		\array: button's equivalent keystroke
	ButX, ButY,	\array: button's coordinates (pixels)
	ButW, ButH,	\array: button's height and width (pixels)
	SpUp,		\addresses of up-button sprites
	SpDn;		\addresses of down-button sprites
int	I,		\scratch index
	MX, MY,		\mouse coordinates (pixels)
	MButton,	\flag: mouse button is pressed
	ButClicked,	\index of button that has been clicked on
	ButDown;	\flag: button is in down (active) position
begin
ShowMouse(false);
for I:= 0, Num-1 do
	DrawSprite(ButX(I), ButY(I), SpUp(I), false);
ClosePixBuf;

CallInt($33, 9, 0, 0, PtrArrow, 0, 0, DataSeg);	\set mouse pointer image
ShowMouse(true);

MButton:= false;
ButClicked:= -1;
while not ChkKey do
	begin
	MX:= GetMousePosition(0);
	MY:= GetMousePosition(1);
	if GetMouseButton then			\mouse button is down
		begin
		if ~MButton then		\mouse button just went down
			begin
			MButton:= true;
			\If clicked on a soft button then make it pending
			for I:= 0, Num-1 do
			    if MX>=ButX(I) & MX<ButX(I)+ButW(I) &
			       MY>=ButY(I) & MY<ButY(I)+ButH(I) then
				   ButClicked:= I;
			end;

		\Mouse button is down. If a soft button is pending then draw it
		\ in up or down position depending on where mouse is pointing.
		if ButClicked >= 0 then
			begin
			I:= ButClicked;
			ButDown:= MX>=ButX(I) & MX<ButX(I)+ButW(I) &
				  MY>=ButY(I) & MY<ButY(I)+ButH(I);
			ShowMouse(false);
			DrawSprite(ButX(I), ButY(I), 
				if ButDown then SpDn(I) else SpUp(I), false);
			ClosePixBuf;
			ShowMouse(true);
			end;
		end
	else	begin			\mouse button is released
		MButton:= false;
		\If mouse is still on pending soft button then execute its
		\ action by pushing its letter into the keyboard buffer
		if ButClicked >= 0 then
			begin		\push command char into keyboard buffer
			if ButDown then InsertKey(ButCh(ButClicked));
			ButClicked:= -1;
			end;
		end;
	Delay(1);
	end;	\while
end;	\DoButtons

\-------------------------------------------------------------------------------

func	SaveChanges;	\Ask if changes should be saved
\ Save changes?		    [X]
\
\ [  Yes  ]  [  No  ]  [Cancel]
\
\User responds with either a keystroke or by pressing a soft button
\Returns keystroke (or equivalent soft button) command
int	I,
	ButCh,		\array: button equivalent keystroke
	ButX, ButY,	\array: button coordinates (pixels)
	ButW, ButH,	\array: button height and width (pixels)
	SpUp(4),	\addresses of up button sprites
	SpDn(4);	\addresses of down button sprites

def	BW=42, BH=14,			\height & width of main buttons (pixels)
	FW=3*(BW/FontWidth+2)+2,	\frame width and height (char cells)
	FH=7,
	LM=(VWidth/FontWidth-FW)/4,	\text left margin (char cells)
	TM=(VHeight/FontHeight-FH)/4,	\text top margin (char cells)
	XL=LM*FontWidth,		\left margin (pixels)
	YT=TM*FontHeight;		\top margin (pixels)
begin
ButCh:= [^y, ^n, Esc, Esc];
ButX:= [XL+12, XL+12+BW+12, XL+12+BW+12+BW+12, XL+FW*FontWidth-14];
ButY:= [YT+5*FontHeight, YT+5*FontHeight, YT+5*FontHeight, YT-18];
ButW:= [BW, BW, BW, 15];
ButH:= [BH, BH, BH, 13];
for I:= 0, 2 do
	begin
	SpUp(I):= addr SpButUp(I,0);
	SpDn(I):= addr SpButDn(I,0);
	end;
SpUp(3):= SpXUp;
SpDn(3):= SpXDn;

ShowMouse(false);
DrawFrame(LM, TM, FW, FH, "Caution!");
ColorFg:= Black;   ColorBg:= Transparent;
LMargin:= LM*FontWidth;
MovePen(LMargin, TM*FontHeight);

DrawText("
  Save changes?");
ShowMouse(true);

loop	begin
	DoButtons(4, ButCh, ButX, ButY, ButW, ButH, SpUp, SpDn);
	if ChkKey then
		begin
		case GetKey of
		  ^y, ^Y:
			return ^y;
		  ^n, ^N:
			return ^n;
		  Esc:
			return Esc
		other	Beep;			\unrecognized command
		end;
	end;
end;	\SaveChanges

\-------------------------------------------------------------------------------

func	PrintScreen;		\Print screen command (except menu bar)
\ Select printer:	[X]
\
\   (o) Laser/Ink
\   ( ) Dot matrix
\
\ [ Print ]  [Cancel]
\
\User responds with either a keystroke or by pressing a soft button. There are
\ 2 normal buttons: Print and Cancel; and 2 radio buttons: Laser and Dot matrix.
\Returns keystroke (or equivalent soft button) command (p or Esc).
def	Pr=5;		\printer device number (2=redirectable; 5=fast and no
			\ "Abort, Retry, Ignore")
int	I,
	ButCh,		\array: button equivalent keystroke
	ButX, ButY,	\array: button coordinates (pixels)
	ButW, ButH,	\array: button height and width (pixels)
	SpUp(5),	\addresses of down button sprites
	SpDn(5),	\addresses of up button sprites
	Printer;	\selected printer: 0=Laser, 1=Dot matrix

def	BW=42, BH=14,	\height and width of normal buttons (pixels)
	FW=2*(BW/FontWidth+4)+4,	\frame width and height (char cells)
	FH=9,
	LM=(VWidth/FontWidth-FW)/4,	\text left margin (char cells)
	TM=(VHeight/FontHeight-FH)/4,	\text top margin (char cells)
	XL=LM*FontWidth,		\left margin (pixels)
	YT=TM*FontHeight;		\top margin (pixels)



proc	EscTxt(Str);	\Send an Esc command to the printer
char	Str;
begin
Chout(Pr, Esc);
Text(Pr, Str);
end;	\EscTxt



proc	Do9Pin;		\Print the current graphic screen on Epson FX-80 printer
\Also works for any standard 24-pin dot matrix printer (IBM Proprinter, ancient
\ Okidata printers, C.Itoh, and DEC "Sixel" printers are not supported). The
\ image is rotated 90 degrees so the top of Screen prints on the left side.
int	L,		\printer line (8 pixels high)
	X, Y,		\coordinates in Screen
	X0,		\X coordinate for base of current print line
	B;		\graphic byte to send to the printer
begin
EscTxt("@");				\set printer to normal settings
EscTxt("A");   Chout(Pr, 8);		\set linefeed pitch to 8/72 inch
					\ (or 8/60 for 24-pin printers)
X0:= 0;
for L:= 0, VWidth/8-1 do
	begin
	EscTxt("K");			\60 dots per inch
	Y:= VHeight-FontHeight-1;	\skip menu bar
	Chout(Pr, Y);   Chout(Pr, Y>>8);
	for Y:= FontHeight+1, VHeight-1 do
		begin
		B:= 0;
		for X:= X0, X0+7 do
			begin
			B:= B << 1;	\top printer pin is MSB
			if ~ChkPix(VWidth-1-X, Y, Black) then B:= B ! $01;
			end;
		Chout(Pr, B);
		end;
	Crlf(Pr);
	X0:= X0 + 8;
	end;

Chout(Pr, FF);
EscTxt("@");				\restore printer to normal settings
end;	\Do9Pin



proc	DoLaser;	\Print current graphic screen on Laser (PCL5) printer
int	X, Y,		\coordinates in Screen
	Y0,		\Y coordinate for base of current byte
	B;		\graphic byte to send to the printer
begin
EscTxt("E");			\set printer to normal settings
EscTxt("*t75R");		\set resolution
EscTxt("*p270x210Y");		\position cursor to center image on page
EscTxt("*r1A");			\start raster graphics at current location

for X:= -(VWidth-1), 0 do
	begin
	EscTxt("*b");		\start graphic scan line
	Intout(Pr, (VHeight-FontHeight-1+7)/8);
	Chout(Pr, ^W);		\skip menu bar but round up to next full byte
	Y0:= FontHeight+1;
	repeat	begin
		B:= 0;
		for Y:= Y0, Y0+7 do	\8 bits in a byte...
			begin
			B:= B << 1;
			if Y < VHeight then
			    if ~ChkPix(-X, Y, Black) then B:= B ! $01;
			end;
		Chout(Pr, B);
		Y0:= Y0 + 8;
		end;
	until Y0 >= VHeight;
	end;

EscTxt("*rB");			\end raster graphics
Chout(Pr, FF);
EscTxt("E");			\restore printer to normal settings
end;	\DoLaser



proc	SetRadios;	\Set radio button sprites
begin
SpUp(2):= if Printer(0)=0 \laser\ then SpRadioOnUp else SpRadioOffUp;
SpDn(2):= if Printer(0)=0 \laser\ then SpRadioOnDn else SpRadioOffDn;
SpUp(3):= if Printer(0)=1 \matrix\ then SpRadioOnUp else SpRadioOffUp;
SpDn(3):= if Printer(0)=1 \matrix\ then SpRadioOnDn else SpRadioOffDn;
end;	\SetRadios



begin	\PrintScreen
Printer:= [0];			\default = Laser = 0
ButCh:= [Esc, ^p, ^l, ^d, Esc];
ButX:= [XL+24+BW+24, XL+24, XL+4*FontWidth, XL+4*FontWidth, XL+FW*FontWidth-14];
ButY:= [YT+7*FontHeight, YT+7*FontHeight, YT+3*FontHeight+1, YT+4*FontHeight+1,
	YT-18];
ButW:= [BW, BW, BW, BW, 15];
ButH:= [BH, BH, FontHeight, FontHeight, 13];
SpUp(0):= addr SpButUp(2,0);
SpDn(0):= addr SpButDn(2,0);
SpUp(1):= addr SpButUp(3,0);
SpDn(1):= addr SpButDn(3,0);
SpUp(4):= SpXUp;
SpDn(4):= SpXDn;

ShowMouse(false);
DrawFrame(LM, TM, FW, FH, "Print");
ColorFg:= Black;   ColorBg:= Transparent;
LMargin:= LM*FontWidth;
MovePen(LMargin, TM*FontHeight);

DrawText("
 Select printer:

       Laser/Ink
       Dot matrix
");

MovePen(XL+7*FontWidth, YT+3*FontHeight+10);		\underline hot keys: L&D
DrawLine(XL+7*FontWidth+4, YT+3*FontHeight+10, Black);
MovePen(XL+7*FontWidth, YT+4*FontHeight+10);
DrawLine(XL+7*FontWidth+4, YT+4*FontHeight+10, Black);
ShowMouse(true);

SetRadios;

loop	begin
	DoButtons(5, ButCh, ButX, ButY, ButW, ButH, SpUp, SpDn);
	if ChkKey then
		begin
		case GetKey of
		  ^p, ^P:
			begin
			DrawCircuit;			\remove dialog box
			ShowMouse(false);
			MovePen(0, FontHeight+1);	\make border lines
			DrawLine(VWidth-1, FontHeight+1, White);
			DrawLine(VWidth-1, VHeight-1, White);
			DrawLine(0, VHeight-1, White);
			DrawLine(0, FontHeight+1, White);
			ClosePixBuf;
			if Printer(0) = 0 then DoLaser else Do9Pin;
			DrawCircuit;			\erase border lines
			ShowMouse(true);
			return ^p;
			end;
		  ^l, ^L:
			begin
			Printer(0):= 0;
			SetRadios;
			end;
		  ^d, ^D:
			begin
			Printer(0):= 1;
			SetRadios;
			end;
		  Esc:
			return Esc
		other	Beep;				\unrecognized command
		end;
	end;	\loop
end;	\PrintScreen

\===============================================================================

proc	Sort(A, N);		\Sort array using the shell sort method
int	A,			\address of array
	N;			\number of elements in array (size)
int	J, Gap, I, JG, T;
begin
Gap:= N/2;
while Gap > 0 do
	begin
	for I:= Gap, N-1 do
		begin
		J:= I - Gap;
		loop	begin
			if J < 0 then quit;
			JG:= J + Gap;
			if StrNCmp(A(J), A(JG), 8) <= 0 then quit;
			T:= A(J);   A(J):= A(JG);   A(JG):= T;
			J:= J - Gap;
			end;
		end;
		Gap:= Gap/2;
	end;
end;	\Sort

\-------------------------------------------------------------------------------

func	GetFileName(Title, FN);	\Prompt user for a file name
char	Title,		\string to display on the top line ("title bar")
	FN;		\string to return file name in (c:/path/xxxxxxxx.lab)
int	FMax,		\index to last file name + 1
	I, J, Ch,	\indexes, character
	In, Out,	\input and output indexes for Buffer
	HaveExt;	\file extension (usually: ".lab") was typed in
char	Ext;		\file name extension string: ".LIB"
def	FilesMax=200;	\maximum number of .LAB file names allowed
int	Names(FilesMax, 8/IntSize);	\2-dimensional char array
def	BufSize=FileNameSize+1;	\maximum number of characters including CR
char	Buffer(BufSize);	\for buffered keyboard

def	FW=6*(1+8+1), FH=19,		\frame width and height (char cells)
	LM=(VWidth/FontWidth-FW)/4,	\text left margin (char cells)
	TM=(VHeight/FontHeight-FH)/4;	\text top margin (char cells)



func	Chin2;			\Buffered keyboard handler
\(DOS routines in graphic mode don't provide a flashing cursor, don't
\ return on Esc, and don't allow the mouse to select a file name.)
int	Ch;



func	GetKey2;
\Non-buffered, non-echoed keyboard input with flashing cursor (in graphic mode).
\The mouse can be used to jam a complete file name into the keyboard's buffer.
int	Cur,		\currently displayed cursor (alternates with space char)
	Cnt;		\count-down timer for cursor flashing speed
int	MX, MY,		\mouse coordinates
	I, J, K, K1, K2, Ch;
char	Str;
int	Col, Row, Col0, Row0;	\column and row of file names; old colum and row
int	ButX, ButY,	\[X] button stuff
	MButton,	\flag: mouse button is pressed
	ButClicked,	\flag: [X] button has been clicked on
	ButDown,	\flag: [X] button is in down (active) position
	FnPend;		\index of pending file name
def	FlashRate=20,	\60ths of a second (for half an on/off cycle)
	CurCh=^_;	\cursor character
begin
Cnt:= FlashRate;
Cur:= CurCh;
Col0:= -1;			\force initial hilite

ButX:= (LM+FW)*FontWidth-14;
ButY:= TM*FontHeight-18;
MButton:= false;
ButClicked:= false;
FnPend:= -1;

loop	begin
	\Display flashing cursor and check for keystroke
	ShowMouse(false);
	DrawChar(Cur);
	ShowMouse(true);
	PenX:= PenX - FontWidth;	\backspace
	if ChkKey then quit;
	Delay(1);
	Cnt:= Cnt - 1;
	if Cnt <= 0 then
		begin			\alternate cursor on/off
		Cnt:= FlashRate;
		Cur:= if Cur=^  then CurCh else ^ ;
		end;

	MX:= GetMousePosition(0);
	MY:= GetMousePosition(1);
	K1:= MY/FontHeight - TM;		\char cell from top margin
	K2:= MX/FontWidth - LM;			\char cell from left margin
	K:= K2/10*FH + K1;			\index to Names

	if GetMouseButton then			\mouse button is down
		begin
		if ~MButton then		\mouse button just went down
			begin
			MButton:= true;
			if MX>=ButX & MX<ButX+15 &
			   MY>=ButY & MY<ButY+13 then		\[X] button
			   	ButClicked:= true;
			if K1>=0 & K1<FH & K2>=0 & K2<FW \within box\ & K<FMax
				then FnPend:= K;	\make file name pending
			end;
		if ButClicked then		\[X] button goes up and down
			begin
			ButDown:= MX>=ButX & MX<ButX+15 &
				  MY>=ButY & MY<ButY+13;
			ShowMouse(false);
			DrawSprite(ButX, ButY, if ButDown then SpXDn else SpXUp,
				false);
			ClosePixBuf;
			ShowMouse(true);
			end;
		end
	else	begin				\mouse button is released
		MButton:= false;
		if ButClicked then		\[X] button
			begin
			ButClicked:= false;
			if ButDown then InsertKey(Esc);
			end;

		\If mouse is still on the pending file name then jam the
		\ file name into the keyboard buffer
		if FnPend >= 0 then		\if there is a pending file name
			begin
			I:= FnPend;
			FnPend:= -1;
			if K1>=0 & K1<FH & K2>=0 & K2<FW \within box\ & K<FMax &
			   K=I then		\ and it's the one at the mouse,
				begin		\ jam file name into keyboard
				In:= 0;		\open keyboard Buffer to delete
						\ any partial name typed in
				Str:= Names(K);	\put int array into char array
				for J:= 0, 7 do	\push name into keyboard buffer
					begin
					Ch:= Str(J) & $7F;
					if Ch = ^  then J:= 8
					else InsertKey(Ch);
					end;
				InsertKey(CR);
				quit;
				end;
			end;
		end;

	\Highlight the file name pointed to by the mouse
	if K1>=0 & K1<FH & K2>=0 & K2<FW \within box\ & K<FMax then
		begin
		Col:= (MX/FontWidth - LM) / 10;
		Row:= MY/FontHeight - TM;
		if Col#Col0 ! Row#Row0 then	\mouse moved to another name
			begin
			if Col0 >= 0 then	\remove green hilite on old name
			    Hilite((Col0*10+LM)*FontWidth, (Row0+TM)*FontHeight,
				10*FontWidth, FontHeight, White);
			Hilite((Col*10+LM)*FontWidth, (Row+TM)*FontHeight, 
				10*FontWidth, FontHeight, Green);
			Col0:= Col;   Row0:= Row;
			end;
		end
	else	begin				\mouse pointer outside of names
		if Col0 >= 0 then		\remove green hilite on old name
		    begin
		    Hilite((Col0*10+LM)*FontWidth, (Row0+TM)*FontHeight,
			10*FontWidth, FontHeight, White);
		    Col0:= -1;
		    end;
		end;
	end;	\loop

ShowMouse(false);
DrawChar(^ );			\make sure flashing cursor is turned off
ShowMouse(true);
PenX:= PenX - FontWidth;	\backspace
return GetKey;
end;	\GetKey2



begin	\Chin2
\If something is in the buffer then return it, otherwise get characters from the
\ keyboard until the Enter key (CR) is struck and then return the first char
if Out = In then		\Buffer is empty
	begin
	In:= 0;   Out:= 0;
	loop	begin
		ShowMouse(false);
		ClosePixBuf;
		ShowMouse(true);
		Ch:= GetKey2;
		case Ch of
		  BS, -LtArrow:	\backspace
			if In > 0 then
				[In:= In - 1;   PenX:= PenX - FontWidth];
		  Esc:	begin
			In:= 0;   Out:= 0;
			Buffer(In):= CR;
			In:= In + 1;
			quit;
			end;
		  CR:	begin
			Buffer(In):= CR;
			In:= In + 1;
			quit;
			end
		other	if Ch>=$20 & Ch<=$7E & In<BufSize-1 then
				begin			\(leave room for the CR)
				ShowMouse(false);
				DrawChar(Ch);		\echo character
				ShowMouse(true);
				Buffer(In):= Ch;
				In:= In + 1;
				end;
		end;
	end;

Ch:= Buffer(Out);
Out:= Out + 1;

ShowMouse(false);
ClosePixBuf;
ShowMouse(true);
return Ch;
end;	\Chin2



proc	ShowFiles;	\Show list of files in a window frame
int	YB, I, K, X, Y;
begin
ShowMouse(false);
DrawFrame(LM, TM, FW, FH, Title);

MovePen((LM+2)*FontWidth, (TM+FH)*FontHeight+6+2);
ColorFg:= Black;   ColorBg:= Transparent;
DrawText("File name: ");

\Draw a little window in the bottom edge of the frame
YB:= (TM+FH)*FontHeight;	\coordinate of top edge of bottom of frame
DrawSprite((LM+13)*FontWidth, YB, SpFrameA, false);
for I:= 0, FileNameSize do
	DrawSprite((LM+14+I)*FontWidth, YB, SpFrameB, false);
DrawSprite((LM+FileNameSize+15)*FontWidth, YB, SpFrameC, false);

\Display file names:
\Each entry is 8 chars wide plus a space on either side, for a total of 10
K:= 0;
for X:= 0, 5 do
	begin
	for Y:= TM, TM+FH-1 do
		begin
		MovePen((LM+X*10+1)*FontWidth, Y*FontHeight);
		if K < FMax then DrawText(Names(K)) else Y:= 30000;
		K:= K + 1;
		end;
	end;
ClosePixBuf;
ShowMouse(true);
end;	\ShowFiles



func	ReadFileNames;
\Read the current directory and store the file names that match Template into
\ Names array. Returns the number of file names stored.
int	I, J, K, Ch;
char	Template,	\type of files to list (limited to file extension here)
	Str,
	DTA(43);
begin	
Template:= "*.LAB ";
\Template:= "*.??? ";	\\debug - tests lots of files
Template(5):= 0;				\terminate string with 0

CallInt($21, $1A00, 0, 0, DTA, 0, DataSeg);	\set Disk Transfer Area address

K:= 0;
\Look up first file name
if CallInt($21, $4E00, 0, $FFF7, Template, 0, DataSeg) & $00FF then return K;

loop	begin
	J:= 0;
	Str:= Names(K);
	loop	begin
		Ch:= DTA(30+J);
		if Ch=^. ! J>=8 then quit;	\beware of blank extensions
		Str(J):= if J=0 then Ch else ToLower(Ch);
		J:= J + 1;
		end;
	for I:= J, 7 do Str(I):= ^ ;
	Str(7):= Str(7) ! $80;
	K:= K + 1;
	if CallInt($21, $4F00) & $00FF then quit;	\look up next file name
	if K >= FilesMax then quit;
	end;
return K;
end;	\ReadFileNames



begin	\GetFileName
FMax:= ReadFileNames;			\read list of file names
Sort(Names, FMax);			\put them into alphabetical order
ShowFiles;				\display 'em

CallInt($33, 9, 0, 0, PtrArrow, 0, 0, DataSeg);		\set mouse pointer image
MovePen((LM+14)*FontWidth, (TM+FH)*FontHeight+6+2);	\move to input window
ColorFg:= Black;   ColorBg:= White;	\(BS requires White, not Transparent)

In:= 0;
Out:= 0;
I:= 0;
HaveExt:= false;
loop	begin
	if I >= FileNameSize then quit;
	Ch:= Chin2;
	if Ch = CR then quit;
	if Ch = ^. then HaveExt:= true;
	FN(I):= Ch;
	I:= I + 1;
	end;
if HaveExt then
	FN(I-1):= FN(I-1) ! $80		\terminate string
else	begin				\add extension if not given
	Ext:= ".LAB";
	for J:= 0, 3 do
		if I+J < FileNameSize then FN(I+J):= Ext(J);
	end;
end;	\GetFileName

\===============================================================================

proc	DoKeyCmd;	\Execute command from keyboard
int	Ch, I, J, K, MX, MY, DX, DY;
char	FN(FileNameSize);
int	In(DataMax),
	Lab(3);		\Label 3 words = 6 characters


	proc	Add(Obj);	\Add an object to the circuit
	int	Obj;
	int	MX, MY;
	begin
	if DataSize >= DataMax then [Beep; return];
	Backup;
	ObjType(DataSize):= Obj;
	MX:= GetMousePosition(0);
	MY:= GetMousePosition(1);
	ObjX(DataSize):= (MX & ~3) - 4;
	ObjY(DataSize):= (MY & ~3) - 8;
	ObjIn1(DataSize):= -1;		\connected to an undefined node number
	ObjIn2(DataSize):= -1;
	ObjOut(DataSize):= true;	\set undefined output to 'high'
	ObjSel(DataSize):= false;
	DataSize:= DataSize + 1;
	DrawCircuit;
	Dirty:= true;
	end;	\Add



	proc	GetLabel(Str, N); \Type in a label
	char	Str;	\address of string to write to (Reserve N bytes)
	int	N;	\maximum number of characters allowed in string
	int	I, Ch,
		Cnt;		\count-down timer for cursor flashing speed
	def	FlashRate=20,	\60ths of a second (for half an on/off cycle)
		CurCh=^_;	\cursor character
	begin
	ShowMouse(false);
	for I:= 0, N-1 do Str(I):= 0;
	ColorFg:= White;   ColorBg:= Transparent;
	Cnt:= FlashRate;
	I:= 0;
	loop	begin
		loop	begin
			\Display flashing cursor and check for keystroke
			DrawChar(CurCh);
			PenX:= PenX - FontWidth;	\backspace
			if ChkKey then quit;
			Delay(1);
			Cnt:= Cnt - 1;
			if Cnt <= 0 then
				begin			\alternate cursor on/off
				Cnt:= FlashRate;
				ColorFg:=if ColorFg=White then Black else White;
				end;
			end;
		ColorFg:= Black;		\erase cursor
		DrawChar(CurCh);
		PenX:= PenX - FontWidth;	\backspace
		ColorFg:= White;

		Ch:= GetKey;
		case Ch of
		  CR, Esc, ^": quit;
		  BS, -LtArrow:
			begin
			if I > 0 then
				begin
				I:= I - 1;
				PenX:= PenX - FontWidth;
				ColorFg:= Black; \overstrike character in black
				DrawChar(Str(I));
				ClosePixBuf;
				ColorFg:= White;
				PenX:= PenX - FontWidth;
				Str(I):= 0;	 \completely eliminate character
				end;
			end
		other	begin
			if I<N & Ch#LF & Ch>0 & Ch<$FF then 
				begin
				Str(I):= Ch;
				I:= I + 1;
				DrawChar(Ch);	\echo character
				ClosePixBuf;
				end;
			end;
		end;
	ShowMouse(true);
	end;	\GetLabel



	proc	GetSaveFileName;
	int	I, Ch;
	begin
	for I:= 0, 7 do			\put current file name into keyboard buf
		begin
		Ch:= FileName(I);
		if Ch = ^. then I:= 8
		else InsertKey(Ch);
		end;
	GetFileName("Save Circuit As", FN);
	end;	\GetSaveFileName


begin	\DoKeyCmd
MX:= GetMousePosition(0);
MY:= GetMousePosition(1);
Ch:= GetKey;
case Ch of
  -UpArrow:
	CallInt($33, 4, 0, MX, MY-4);
  -DnArrow:
	CallInt($33, 4, 0, MX, MY+4);
  -LtArrow:
	CallInt($33, 4, 0, MX-4, MY);
  -RtArrow:
	CallInt($33, 4, 0, MX+4, MY);

  ^p, ^P:							\Print
	begin
	PrintScreen;
	DrawCircuit;			\(in case print was canceled)
	end;
  ^h, ^H, -Func1:
	begin							\Help
	ShowHelp;
	DrawCircuit;
	end;
  ^?, -Func2:
	begin							\About
	ShowAbout;
	DrawCircuit;
	end;
  ^s, ^S:
	begin							\Save
	GetSaveFileName;
	if FN(0) # ^. then SaveData(FN); \(don't beep if Esc was entered)
	DrawCircuit;
	end;
  ^l, ^L:
	begin							\Load
	if Dirty then
		begin
		Ch:= SaveChanges;
		if Ch = ^y then
			begin
			GetSaveFileName;
			if FN(0) = ^. then Ch:= Esc else SaveData(FN);
			end;
		end;
	if Ch # Esc then		\Ch = l, L, n, y
		begin
		GetFileName("Load Circuit File", FN);
		if FN(0) # ^. then [Backup; LoadData(FN)];
		end;
	CalcCircuit;
	DrawCircuit;
	end;
  ^i, ^I, -Insert:						\Insert
	begin
	GetFileName("Insert Circuit File", FN);
	if FN(0) # ^. then
		begin
		Backup;
		InsertData(FN);

		\Put selection box around inserted objects by finding
		\ minimum and maximum locations of inserted objects
		SelBox0X:= VWidth;
		SelBox0Y:= VHeight;
		SelBox1X:= 0;
		SelBox1Y:= 0;
		for I:= 0, DataSize-1 do
		    if ObjSel(I) then
			begin
			if ObjX(I) < SelBox0X then SelBox0X:= ObjX(I);
			if ObjY(I) < SelBox0Y then SelBox0Y:= ObjY(I);
			if ObjX(I) > SelBox1X then SelBox1X:= ObjX(I);
			if ObjY(I) > SelBox1Y then SelBox1Y:= ObjY(I);
			end;
		SelBox0X:= SelBox0X - 4;	\make box slightly bigger
		SelBox0Y:= SelBox0Y - 4;	\ and on a grid point
		SelBox1X:= SelBox1X + 30;
		SelBox1Y:= SelBox1Y + 20;

		\Move upper-left corner of selection box to mouse pointer
		DX:= (MX&~3) - SelBox0X;
		DY:= (MY&~3) - SelBox0Y;
		for I:= 0, DataSize-1 do  	\move contents
		    if ObjSel(I) then
			begin
			ObjX(I):= ObjX(I) + DX;
			ObjY(I):= ObjY(I) + DY;
			end;
		SelBox0X:= SelBox0X + DX; 	\move box
		SelBox0Y:= SelBox0Y + DY;
		SelBox1X:= SelBox1X + DX;
		SelBox1Y:= SelBox1Y + DY;
		end;
	CalcCircuit;
	DrawCircuit;
	end;

  ^K:	begin							\Kill all data
	if Dirty then
		begin
		Ch:= SaveChanges;
		if Ch = ^y then
			begin
			GetSaveFileName;
			if FN(0) = ^. then Ch:= Esc else SaveData(FN);
			end;
		end;
	if Ch # Esc then		\Ch = K, n, y
		[Backup; DelSelBox; DataSize:= 0; Dirty:= false];
	DrawCircuit;			\(show the menu bar)
	end;

  ^a, ^A: Add(AND);						\Add an object
  ^o, ^O: Add(OR);
  ^n:	  Add(NOT);
  ^N:	  Add(NOT2);
  ^b, ^B: Add(BUT);
  ^d:	  Add(NAND);
  ^D:	  Add(NAND2);
  ^r:	  Add(NOR);
  ^R:	  Add(NOR2);
  ^x, ^X: Add(XOR);
  ^q, ^Q: Add(LATCH);
  ^f, ^F: Add(FLIP);
  ^e, ^E: Add(NODE);
  ^c, ^C: Add(CLK);
  ^t, ^T: Add(TERM);

  ^":	begin							\Add a label
	if DataSize < DataMax then
		begin
		Backup;
		ObjType(DataSize):= LABEL;

		MX:= GetMousePosition(0);
		MY:= GetMousePosition(1);
		PenX:= (MX & ~3);
		PenY:= ((MY+2) & ~3) - 8 - 1;	\start at base line & up a pixel
		ObjX(DataSize):= PenX;
		ObjY(DataSize):= PenY;
		GetLabel(Lab, 6);
		ObjOut(DataSize):= Lab(0);
		ObjIn1(DataSize):= Lab(1);
		ObjIn2(DataSize):= Lab(2);

		ObjSel(DataSize):= false;
		DataSize:= DataSize + 1;
		DrawCircuit;
		Dirty:= true;
		end
	else	Beep;
	end;
  ^g, ^G:
	begin							\Grid on/off
	HaveGrid:= ~HaveGrid;
	DrawCircuit;
	end;
  ^m, ^M:							\DeMorganize
	begin
	if SelBox1Y then		\selection box exists
		begin
		for I:= 0, DataSize-1 do
		    if ObjSel(I) then
			begin
			case ObjType(I) of
			  NAND:	ObjType(I):= NAND2;
			  NAND2:ObjType(I):= NAND;
			  NOR:	ObjType(I):= NOR2;
			  NOR2:	ObjType(I):= NOR;
			  NOT:	ObjType(I):= NOT2;
			  NOT2:	ObjType(I):= NOT
			other	[];
			Dirty:= true;
			end;
		DrawCircuit;
		end
	else	Beep;			\must have box
	end;
  ^v, ^V:							\InVert
	begin
	if SelBox1Y then		\selection box exists
		begin
		for I:= 0, DataSize-1 do
		    if ObjSel(I) then
			begin
			ObjY(I):= SelBox0Y + (Selbox1Y - ObjY(I)) - 16;
			case ObjType(I) of
			  AND, OR, NAND, NOR, XOR, NAND2, NOR2:
				begin	\swap inputs
				J:= ObjIn1(I);
				ObjIn1(I):= ObjIn2(I);
				ObjIn2(I):= J;
				end
			other	[];
			Dirty:= true;
			end;
		DrawCircuit;
		end
	else	Beep;			\must have box
	end;

  -Delete, BS, ^X-Ctrl\cut\:				       \Delete object(s)
    if DataSize > 0 then
	begin
	Backup;
	if SelBox1Y then		 \there is a selection box
		begin
		for I:= 0, DataSize-1 do \delete all objects in it
		    if ObjSel(I) then DelObj(I);
		DelSelBox;		  \also delete the selection box
		DelNULLs;
		end
	else	begin			\delete last object entered
		DataSize:= DataSize - 1;
		DelObj(DataSize);
		end;
	Dirty:= true;
	if DataSize = 0 then Dirty:= false;
	CalcCircuit;
	DrawCircuit;
	end;
  CR, ^V-Ctrl\paste\:						\Duplicate
    \Add a duplicate copy of the objects in the selection box
    if SelBox1Y then			\there is a selection box
	loop begin	\used to abort if overflow; doesn't actually loop
	Backup;
	J:= DataSize;
	for I:= 0, DataSize-1 do
	    if ObjSel(I) then
		begin
		ObjType(J):= ObjType(I);
		ObjX(J):= ObjX(I) + 4;	\offset duplicate slightly so that it
		ObjY(J):= ObjY(I) + 4;	\ is visible
		ObjIn1(J):= ObjIn1(I);
		ObjIn2(J):= ObjIn2(I);
		ObjOut(J):= ObjOut(I);
		In(J):= I;		\record original index of this duplicate
		ObjSel(J):= true;	\select duplicate and
		ObjSel(I):= false;	\ deselect original

		J:= J + 1;
		if J >= DataMax then
			[Beep;   quit];	\(don't duplicate partial data)
		end;

        \At this point the inputs to the duplicated objects are connected to
        \ the original output objects. Now transfer these input connections to
        \ any output objects that were also duplicated.

	for I:= DataSize, J-1 do	\for the duplicated objects...
	    if ObjType(I) # LABEL then
		begin
		for K:= DataSize, J-1 do
			begin
			if ObjIn1(I) = In(K) then  \output has been copied too
				ObjIn1(I):= K;	   \xfer connection to copy
			if ObjIn2(I) = In(K) then
				ObjIn2(I):= K;
			end;
		end;
	DataSize:= J;

	Dirty:= true;
	DrawCircuit;
	quit;
	end
   else Beep;		\must have box

  ^u, ^U:							\Undo
	Undo;

  Esc, ^C-Ctrl:							\Exit
	begin
	Ch:= ^ ;	\trick-o change-o!
	if Dirty then
		begin
		Ch:= SaveChanges;
		if Ch = ^y then
			begin
			GetSaveFileName;
			if FN(0) = ^. then Ch:= Esc else SaveData(FN);
			end;
		end;

	if Ch # Esc then		\Ch = space, y, n
		begin
		SaveData("_BACKUP_.LAB");
		ShowMouse(false);
		SetVid($03);		\restore normal text mode
		exit;
		end;
	DrawCircuit;
	end

other	if Ch>=^0 & Ch<=^9 then				\clock speed control
		begin
		I:= Ch - ^0;
		ClkSpeed:= I * I;
		ClkCtr:= ClkSpeed;
		FlipClks;
		end
	else	Beep;					\unrecognized command
end;	\DoKeyCmd

\===============================================================================

proc	DoMenuCmd;	\Execute commands from drop-down menus
int	Command,	\command number (1..N) from DoMenu
	MBut,		\flag: mouse button state upon entry to procedure
	MenusOpen,	\flag: menus have dropped down
	MenuWidth0,	\width of previous menu (chars)
	Y00;		\vertical position of previous menu entry (char lines)


func	DoMenu(X0, TWidth, MenuWidth, MenuHeight, Strings);
\Routine to handle a single drop-down menu. Returns command number.
int	X0,		\horizontal position of title incl leading space (chars)
	TWidth,		\title's width (in chars) including surrounding spaces
	MenuWidth, MenuHeight,	\menu's dimensions in character cells
	Strings;	\array of menu legends (has Height number of entries)
int	Cmd,		\command number returned; = 0..MenuHeight (0=none)
	MX, MY,		\mouse coordinates (in character cells)
	N;		\line number (= command number) in menu
char	S;		\scratch string (to access separator line)


	proc	DrawMenu;	\Display the menu
	int	J,	\index
		XL,	\coordinate of left edge of menu frame (pixels)
		XR,	\coordinate of right edge (pixels)
		YT,	\coordinate of top edge (pixels)
		YB;	\coordinate of bottom edge (pixels)
	begin
	ShowMouse(false);
	ColorFg:= Black;   ColorBg:= White;
	for J:= 1, MenuHeight do		\display all strings in the list
		begin
		MovePen(X0*FontWidth, J*FontHeight+2);
		DrawRawText(Strings(J-1));
		end;

	XL:= X0*FontWidth-1;
	XR:= (X0+MenuWidth)*FontWidth;
	YT:= FontHeight;
	YB:= (MenuHeight+1)*FontHeight+1;

	MovePen(XL, YT);			\draw borders
	DrawLine(XR, YT, Black);
	MovePen(XL, YT+1);
	DrawLine(XR, YT+1, White);

	DrawLine(XR, YB+1, White);
	DrawLine(XL, YB+1, White);
	DrawLine(XL, YT+1, White);
	MovePen(XR+1, YT);
	DrawLine(XR+1, YB+2, Green);		\shadow
	DrawLine(XL, YB+2, Green);

	for J:= 1, 3 do				\outline menu with 3 black lines
		begin
		MovePen(XR+1+J, YT+1);
		DrawLine(XR+1+J, YB+2+J, Black);
		DrawLine(XL-J, YB+2+J, Black);
		DrawLine(XL-J, YT+2, Black);
		end;

	ClosePixBuf;
	ShowMouse(true);
	end;	\DrawMenu


begin	\DoMenu
Cmd:= 0;
MX:= GetMousePosition(0)/FontWidth;
MY:= (GetMousePosition(1)-1)/FontHeight;	\menu entries are down 2 pixels;
						\ -1 is a compromise with titles
if ~MenusOpen then
    begin
    if MY#0 ! MX<X00 ! MX>=X00+TWidth0 then
	begin				\mouse is not on previous title
	if X00 >= 0 then
		begin			\previous title was highlighted--undo it
		Hilite(X00*FontWidth, 0, TWidth0*FontWidth, FontHeight, White);
		X00:= -1;		\indicate highlight is gone
		end;
	end;
    if MY=0 & MX>=X0 & MX<X0+TWidth then \mouse is on current (polled) title
	begin
	if X0 # X00 then
		begin			\this is the first time so hilite title
		Hilite(X0*FontWidth, 0, TWidth*FontWidth, FontHeight, Green);
		X00:= X0;   TWidth0:= TWidth;	\record title that is hilighted
		end;
	end;
    end;


if MBut then
	begin
	if ~OldMouseButton then
		begin
		if MY=0 & MX>=1 & MX<26+6 then	\mouse is in title area
			begin
			MenusOpen:= true;
			X00:= -1;		\force menu to be drawn
			end;
		end;
	end;


if MenusOpen then
    begin
    if MY=0 & MX>=X0 & MX<X0+TWidth then
	begin					\mouse is on polled menu title
	if X0 # X00 then
		begin				\first time on this menu title
		if X00 >= 0 then
			begin			\previous menu is displayed
			DrawCircuit;		\erase previous menu
			X00:= -1;		\indicate previous menu is gone
			Y00:= -1;		\indicate highlight is gone
			end;
		Hilite(X0*FontWidth, 0, TWidth*FontWidth, FontHeight, Green);
		X00:= X0;			\record title that is hilighted
		DrawMenu;
		end;
	end;

    if X0 = X00 then				\poll entries in current menu...
      for N:= 1, MenuHeight do
	begin
	if MY#Y00 ! MX<X0 ! MX>=X0+MenuWidth0 then
		begin				\mouse is not on previous entry
		if Y00 >= 0 then		\previous entry was highlighted
			begin			\undo previous highlight
			Hilite(X00*FontWidth, Y00*FontHeight+2,
				MenuWidth0*FontWidth, FontHeight, White);
			Y00:= -1;		\indicate highlight is gone
			end;
		end;
	if MY=N & MX>=X0 & MX<X0+MenuWidth then \mouse is on current entry N
		begin
		if N # Y00 then			\hasn't been highlighted yet
			begin
			S:= Strings(N-1);	\don't hilite if separator line
			if S(2) # ^ then
			    Hilite(X0*FontWidth, N*FontHeight+2,
			    	MenuWidth*FontWidth, FontHeight, Green);
			Y00:= N;		\record entry that is hilighted
			MenuWidth0:= MenuWidth;
			end;
		if ~MBut & OldMouseButton then \mouse button released
			begin
			Cmd:= N;
			MenusOpen:= false;	\erase menu
			DrawCircuit;
			X00:= -1;		\indicate highlight is gone
			Y00:= -1;
			end;
		end;

	if MY>MenuHeight ! MX<X0 ! MX>=X0+MenuWidth then	\not inside menu
	    if (MBut | OldMouseButton) then    \mouse button released or pressed
		begin
		MenusOpen:= false;	       \erase menu
		DrawCircuit;
		X00:= -1;		       \indicate highlight is gone
		Y00:= -1;
		end;

	if ChkKey then				\any keystroke aborts
		begin
		MenusOpen:= false;		\erase menu
		DrawCircuit;
		X00:= -1;			\indicate highlight is gone
		Y00:= -1;
		end;
	end;
    end;	\MenusOpen

return Cmd;
end;	\DoMenu



begin	\DoMenuCmd
MenusOpen:= false;
Y00:= -1;
repeat	begin
	MBut:= GetMouseButton;

	Command:= DoMenu(1, 6, 18, 7,
		["  Load...    l     ",
		 "  Insert...  i     ",
		 "  Save As... s     ",
		 " ",
		 "  Print...   p     ",
		 " ",
		 "  Exit       Esc   "] );
	case Command of			\convert menu command to a key command
	  1:	InsertKey(^l);
	  2:	InsertKey(^i);
	  3:	InsertKey(^s);
	  5:	InsertKey(^p);
	  7:	InsertKey(Esc)
	other	[];

	Command:= DoMenu(7, 6, 18, 8,
		["  Undo       u     ",
		 " ",
		 "  Duplicate  ^Q   ",
		 "  Invert     v     ",
		 "  Delete     Del   ",
		 "  Kill All   K     ",
		 " ",
		 "  DeMorgan   m     "] );
	case Command of
	  1:	InsertKey(^u);
	  3:	InsertKey(CR);
	  4:	InsertKey(^v);
	  5:	InsertKey(Delete<<8);
	  6:	InsertKey(^K);
	  8:	InsertKey(^m)
	other	[];

	Command:= DoMenu(13, 6, 13, 1,
		if HaveGrid then
		["  Grid   g   "]
		else
		["   Grid   g   "] );
	case Command of
	  1:	InsertKey(^g)
	other	[];

	Command:= DoMenu(19, 7, 18, 16,
		["  AND        a     ",
		 "  OR         o     ",
		 "  NOT        n N   ",
		 "  NAND       d D   ",
		 "  NOR        r R   ",
		 "  XOR        x     ",
		 " ",
		 "  Flip Flop  f     ",
		 "  Latch      q     ",
		 " ",
		 "  Node       e     ",
		 "  Terminal   t     ",
		 "  Button     b     ",
		 "  Clock      c     ",
		 " ",
		 "  Label      ^"     "] );
	case Command of
	  1:	InsertKey(^a);
	  2:	InsertKey(^o);
	  3:	InsertKey(^n);
	  4:	InsertKey(^d);
	  5:	InsertKey(^r);
	  6:	InsertKey(^x);
	  8:	InsertKey(^f);
	  9:	InsertKey(^q);
	 11:	InsertKey(^e);
	 12:	InsertKey(^t);
	 13:	InsertKey(^b);
	 14:	InsertKey(^c);
	 16:	InsertKey(^")
	other	[];

	Command:= DoMenu(26, 6, 13, 2,
		["  Help   F1   ",
		 "  About  F2   "] );
	case Command of
	  1:	InsertKey(^h);
	  2:	InsertKey(^?)
	other	[];

	OldMouseButton:= MBut;
	end;
until ~MenusOpen;
end;	\DoMenuCmd

\===============================================================================

proc	DnD;		\Drag 'n drop gates and connections, also [X] button
int	I, L,		\general index
	J,		\index to object to be grabbed and dragged
	K,		\index to object whose input connection is being dragged
	F,		\general flag
	T,		\temporary scratch
	GSB,		\flag: Grabbed Selection Box
	DX, DY,		\amount to move selection box and its contents
	MX, MY,		\mouse coordinates (pixels)
	MX0, MY0,	\previous mouse coordinates (pixels)
	MButton,	\flag: mouse button is pressed
	MPtr, MPtr0,	\mouse pointer type (arrow or cross)
	OX, OY,		\dragged object's (or connection's) coordinates
	OX0, OY0,	\dragged connection's previous coordinates
	OffX, OffY,	\offset of mouse ptr from sprite when object is grabbed
	ObjIn,		\pointer to ObjIn1(K) or ObjIn2(K)
	OYOff,		\Y coordinate offset to input connection (4, 8 or 12)
	SelOffX, SelOffY, \offset of mouse ptr from selection box it is grabbed
	SelPend,	\flag: selection box is being created (expanded/pending)
	W, H,		\width and height of object (pixels)
	ButClicked,	\flag: exit [X] button has been clicked on
	ButDown;	\flag: exit [X] button is in down (active) position
begin
MPtr0:= -1;
MX0:= -1;   MY0:= -1;
MButton:= false;
ButClicked:= false;
J:= -1;			\nothing is grabbed yet
K:= -1;
GSB:= false;
DelSelBox;		\there is no selection box
SelPend:= false;
X00:= -1;
OldMouseButton:= GetMouseButton;

for I:= 0, DataMax-1 do
	begin
	Clock(I):= true;
	ObjSel(I):= false;
	end;
ClkSpeed:= 4*4;		\default clock speed
ClkCtr:= ClkSpeed;

CalcCircuit;
DrawCircuit;
ShowMouse(true);

Backup;
loop	begin
	MX:= GetMousePosition(0);
	MY:= GetMousePosition(1);
	if MX#MX0 ! MY#MY0 then
		begin
		MX0:= MX;   MY0:= MY;
		ColorFg:= Black;   ColorBg:= White;
		MovePen(75*FontWidth, 0);	\also see DrawMenuBar
		ShowMouse(false);
		DrawNum(MX/4);   DrawChar(^,);
		DrawNum(MY/4);   DrawText("    ");
		ShowMouse(true);
		end;

	\Set mouse pointer image depending on whether it's on an input or not
	MPtr:= PtrArrow;
	for I:= 0, DataSize-1 do
	    if MX>=ObjX(I)-1 & MX<ObjX(I)+3 & MY>=ObjY(I)+2 & MY<ObjY(I)+15 &
	          ~ObjSel(I) & ObjType(I)#NULL & ObjType(I)#BUT &
	          ObjType(I)#CLK & ObjType(I)#LABEL then MPtr:= PtrCross;
	if Mptr # MPtr0 then
		begin
		if MPtr = PtrArrow then
			CallInt($33, 9, 0, 0, MPtr, 0, 0, DataSeg)
		else	CallInt($33, 9, 5, 5, MPtr, 0, 0, DataSeg);
		MPtr0:= MPtr;
		end;

	if J<0 & K<0 & ~GSB & ~SelPend then DoMenuCmd;
	if ChkKey then DoKeyCmd;

	if GetMouseButton then
		begin
		if ~MButton then
\----------------------- Mouse button just went down (was clicked) -------------
			begin
			MButton:= true;

			\If selection box exists and mouse is in it then grab it
			GSB:= false;		\assume mouse is not in the box
			if MX>=SelBox0X & MX<SelBox1X &
			   MY>=SelBox0Y & MY<SelBox1Y then
				begin
				Backup;
				GSB:= true;	\selection box is grabbed
				SelOffX:= MX - SelBox0X;	\get offset to
				SelOffY:= MY - SelBox0Y;	\ mouse pointer
				end;

			J:= -1;			    \assume no object is grabbed
			if ~GSB then
			  for I:= 0, DataSize-1 do  \get index of grabbed object
			    begin
			    W:= ObjWidth(ObjType(I));
			    if W < 8 then W:= 8;	\for NODE and TERM
			    H:= ObjHeight(ObjType(I));
			    if MX>=ObjX(I)+3 & MX<ObjX(I)+W &	\3 = input zone
			       MY>=ObjY(I)+(17-H)>>1 & MY<ObjY(I)+(17+H)>>1 then
				begin
				Backup;
				J:= I;	  	  \save index of grabbed object
				OffX:= MX - ObjX(I);	\offset to mouse pointer
				OffY:= MY - ObjY(I);

			        if ObjType(I) = BUT then \flip state of buttons
				    begin
				    \Save initial state of possible clock inputs
				    for L:= 0, DataMax-1 do Clock(L):=ObjOut(L);
				    ObjOut(I):= ~ObjOut(I);	\flip state
				    CalcCircuit;		\show result
				    DrawCircuit;
				    end;
				end;
			    end;

			K:= -1;			 \assume no input is grabbed
			if ~GSB then
			  for I:= 0, DataSize-1 do \get index of grabbed input
			    if MX>=ObjX(I)-1 & MX<ObjX(I)+3 then
			      if ObjType(I)#BUT & ObjType(I)#CLK &
			         ObjType(I)#LABEL then
				begin
				if MY>=ObjY(I)+2 & MY<ObjY(I)+8 then
					begin
					Backup;
			       		K:= I;	 \save index of grabbed input
			       		ObjIn:= addr ObjIn1(K);
					OYOff:= 4;		\near top
			       		if ObjType(K)=NOT ! ObjType(K)=NODE !
			       		   ObjType(K)=TERM! ObjType(K)=NOT2 then
			       			OYOff:= 8;	\midway down
			       		OX0:= -1; \force screen update
			       		end;
				if MY>=ObjY(I)+8 & MY<ObjY(I)+15 then
					begin
					Backup;
			       		K:= I;	 \save index of grabbed input
					ObjIn:= addr ObjIn2(K);
					OYOff:= 12;		\near bottom
			       		if ObjType(K)=NOT ! ObjType(K)=NODE !
			       		   ObjType(K)=TERM! ObjType(K)=NOT2 then
			       			begin
			       			ObjIn:= addr ObjIn1(K);
			       			OYOff:= 8;	\midway down
			       			end;
			       		OX0:= -1; \force screen update
			       		end;
				end;

			\If nothing has been grabbed then start a selection box
			if J<0 & K<0 & ~GSB then
				begin
				DelSelBox;		\delete any existing box
				SelBox0X:= MX & ~3;	\grid snap
				SelBox0Y:= MY & ~3;
				SelBox1X:= MX & ~3;
				SelBox1Y:= MY & ~3;
				SelPend:= true;
				DrawCircuit;		\erase old box, if any
				end;

			if MX>=VWidth-15 & MY<13 then	\clicked on [X] button
				ButClicked:= true;
			end;

\--------------- Mouse button is still down ------------------------------------
		if J >= 0 then			\an object has been grabbed
			begin			\drag it
			OX:= ((MX-OffX) & ~3);	\object's coordinates snapped to
			OY:= ((MY-OffY) & ~3);	\ nearest 4x4 grid point
			if ObjType(J) = LABEL then OY:= OY - 1;	\up a pixel

			if OX#ObjX(J) ! OY#ObjY(J) then
				begin		\object moved so update screen
				ObjX(J):= OX;
				ObjY(J):= OY;
				DrawCircuit;	\(object can drag connections)
				Dirty:= true;
				end;
			end;

		if K >= 0 then			\object's input has been grabbed
			begin			\drag the connection
			OX:= (MX & ~3);		\4x4 grid snap
			OY:= (MY & ~3);
			if OX#OX0 ! OY#OY0 then
				begin		\connection has moved so
				OX0:= OX;	\ update screen
				OY0:= OY;
				DrawCircuit;
				ShowMouse(false);
				DrawConn(ObjX(K), ObjY(K)+OYOff, OX, OY, true);
				ShowMouse(true);
				end;
			end;

		\If nothing has been grabbed then expand the selection box
		if SelPend & SelBox0Y then	\nothing has been grabbed
			begin			\ and box is not deleted
			OX:= (MX & ~3);		\4x4 grid snap
			OY:= (MY & ~3);
			if OX#SelBox1X ! OY#SelBox1Y then
				begin		\expand selection box
				SelBox1X:= OX;
				SelBox1Y:= OY;
				DrawCircuit;	\display the resulting box
				end;
			end;

		if GSB then
			begin
			\Add offset to all selected objects in the box
			OX:= ((MX-SelOffX) & ~3); \box's coordinates snapped to
			OY:= ((MY-SelOffY) & ~3); \ nearest 4x4 grid point
			DX:= OX - SelBox0X;	  \amount to move box & contents
			DY:= OY - SelBox0Y;
			if DX ! DY then
				begin
				for I:= 0, DataSize-1 do  \move contents
				    if ObjSel(I) then
					begin
					ObjX(I):= ObjX(I) + DX;
					ObjY(I):= ObjY(I) + DY;
					end;
				SelBox0X:= SelBox0X + DX; \move box
				SelBox0Y:= SelBox0Y + DY;
				SelBox1X:= SelBox1X + DX;
				SelBox1Y:= SelBox1Y + DY;
				DrawCircuit;		  \show the result
				Dirty:= true;
				end;
			end;

		if ButClicked then
			begin
			ButDown:= MX>=VWidth-15 & MY<13;
			ShowMouse(false);
			DrawSprite(VWidth-15, 0, 
				if ButDown then SpXDn else SpXUp, false);
			ClosePixBuf;
			ShowMouse(true);
			end;
		end

\------- Mouse button is released ----------------------------------------------
	else	begin
		MButton:= false;
		J:= -1;				\release any grabbed object
		GSB:= false;

		if K >= 0 then			\connection was being dragged
			begin
			for I:= 0, DataSize-1 do \get index of output object
			    begin
			    W:= ObjWidth(ObjType(I));
			    if W < 8 then W:= 8;	\for NODE and TERM
			    H:= ObjHeight(ObjType(I));
			    if OX>=ObjX(I)+3 & OX<ObjX(I)+W+3 &	\3 = I/O zones
			       OY>=ObjY(I)+(17-H)>>1 & OY<ObjY(I)+(17+H)>>1 &
			       ObjType(I)#LABEL then
				ObjIn(0):= I;	\connect to output
			    end;
			K:= -1;			\connection has been handled
			CalcCircuit;		\show result (even if connection
			DrawCircuit;		\ failed, because it snaps back)
			Dirty:= true;
			end;

		if SelPend then
			begin
			SelPend:= false;
			\Put corners in to proper order
			if SelBox1X < SelBox0X then
				begin
				T:= SelBox1X;
				SelBox1X:= SelBox0X;
				SelBox0X:= T;
				end;
			if SelBox1Y < SelBox0Y then
				begin
				T:= SelBox1Y;
				SelBox1Y:= SelBox0Y;
				SelBox0Y:= T;
				end;

			\Select all objects whose center is in the selection box
			F:= false;
			for I:= 0, DataSize-1 do
			    begin
			    ObjSel(I):= false;
			    W:= ObjWidth(ObjType(I));
			    if W < 8 then W:= 8;	\for NODE and TERM
			    W:= ObjX(I) + W>>1;
			    H:= ObjY(I) + 17>>1;
			    if W>=SelBox0X & W<=SelBox1X &
			       H>=SelBox0Y & H<=SelBox1Y then
				begin
				F:= true;
				ObjSel(I):= true;
				end;
			    end;

			if ~F then		\there was nothing in the box
				begin		\ so delete it
				DelSelBox;
				DrawCircuit;	\erase box from screen
				end;
			end;

		if ButClicked then
			begin
			ButClicked:= false;
			if ButDown then InsertKey(Esc);
			ShowMouse(false);	\draw button in up position
			DrawSprite(VWidth-15, 0, SpXUp, false);
			ClosePixBuf;
			ShowMouse(true);
			end;

		if ClkSpeed > 0 then
		    begin
		    ClkCtr:= ClkCtr - 1;
		    if ClkCtr <= 0 then
			begin
			ClkCtr:= ClkSpeed;
			FlipClks;
			end;
		    end;
		end;

	Delay(2);
	end;	\loop
end;	\DnD

\===============================================================================

proc	LoadPointer(X0, Y0, Ptr);	\Load mouse pointer image
int	X0, Y0,		\coordinates in Screen to get pointer from
	Ptr;		\pointer image array to load
int	Xor,		\xor portion of Ptr array (the second half)
	S, I, X, Y, C;
begin
Xor:= Ptr + 2*16;		\point to 2nd half of integer array
I:= 0;
for Y:= Y0, Y0+16-1 do
	begin
	Ptr(I):= $FFFF;
	Xor(I):= $0000;
	S:= $8000;
	for X:= X0, X0+16-1 do
		begin
		C:= Screen(0, X+Y*320);
		if C # Black then
			Ptr(I):= Ptr(I) & ~S;
		if C = White then
			Xor(I):= Xor(I) ! S;
		S:= S >> 1;
		end;
	I:= I + 1;
	end;
end;	\LoadPointer



proc	LoadSprite(X0, Y0, W, H, Sp);	\Load a sprite data array from Screen
int	X0, Y0,		\coordinates in Screen to get sprite from
	W, H;		\width and height of sprite (pixels)
char	Sp;		\sprite to load
int	K, X, Y;
begin
Sp(0):= W;
Sp(1):= H;
K:= 2;
for Y:= Y0, Y0+H-1 do
    for X:= X0, X0+W-1 do
	begin
	Sp(K):= Screen(0, X+Y*320);
	K:= K + 1;
	end;
end;	\LoadSprite



proc	LoadSprites;		\Load sprite images from .BMP file into Screen
int	Hand,		\file handle
	W, H,		\dimensions of entire image in .BMP file (pixels)
	R, G, B,	\red, green, blue
	X, Y,		\screen coordinates (pixels)
	Y320,		\Y * 320
	I, T;		\indexes and tempory scratch
begin
\Read in a 256-color .BMP file
Trap(false);			\don't trap "file not found" error; we'll do it
Hand:= FOpen("LOGICLAB.BMP", 0);\open file for input
if Geterr \#0\ then Exit("LOGICLAB.BMP not found");
Trap(Debug);
FSet(Hand, ^I);			\set device 3 to handle
OpenI(3);

for Y:= 0, 17 do X:= Chin(3);	\skip unused header info
W:= Chin(3) + Chin(3)<<8;	\0..32764 (ample range)
for Y:= 0, 2-1 do X:= Chin(3);	\skip
H:= Chin(3) + Chin(3)<<8;	\0..32767 (ample range)
for Y:= 24, 53 do X:= Chin(3);	\skip

port($3C8):= 0;			\set color registers
for I:= 0, 255 do
	begin
	B:= Chin(3)>>2;
	G:= Chin(3)>>2;
	R:= Chin(3)>>2;
	T:= Chin(3);
	port($3C9):= R;
	port($3C9):= G;
	port($3C9):= B;
	end;

\Load .BMP image into Screen
for Y:= -(H-1), 0 do		\.BMP files are upside down (!)
	begin
	Y320:= 320 * -Y;
	for X:= 0, W-1 do
		Screen(0, X+Y320):= Chin(3);
	end;
FClose(Hand);			\close handle so it can be used again

LoadSprite( 0,  0, 22, 17, SpAND);
LoadSprite( 0, 20, 22, 17, SpOR);
LoadSprite( 0, 40, 22, 17, SpNOT);
LoadSprite( 0, 60, 22, 17, SpBUT);
LoadSprite( 0, 80, 22, 17, SpBUTOn);
LoadSprite(40,  0, 26, 17, SpNAND);
LoadSprite(40, 20, 26, 17, SpNOR);
LoadSprite(40, 40, 26, 17, SpXOR);
LoadSprite(80,  0, 23, 17, SpLATCH);
LoadSprite(80, 20, 23, 17, SpFLIP);
LoadSprite(80, 60, 22, 17, SpNODE);
LoadSprite(80, 40, 23, 17, SpCLK);
LoadSprite(40, 60, 22, 17, SpTERM);
LoadSprite(40, 80, 22, 17, SpTERMOn);
LoadSprite(280,20, 26, 17, SpNAND2);
LoadSprite(280, 0, 26, 17, SpNOR2);
LoadSprite(280,40, 22, 17, SpNOT2);

LoadSprite(120,  0, 15, 13, SpXUp);
LoadSprite(140,  0, 15, 13, SpXDn);
LoadSprite(140, 20,  9,  9, SpRadioOffUp);
LoadSprite(140, 40,  9,  9, SpRadioOnUp);
LoadSprite(150, 20,  9,  9, SpRadioOffDn);
LoadSprite(150, 40,  9,  9, SpRadioOnDn);

LoadSprite(160, 38, 12, 30, SpFrame1);
LoadSprite(173, 38,  6, 30, SpFrame2);
LoadSprite(201, 38, 12, 30, SpFrame3);
LoadSprite(160, 25, 12, 12, SpFrame4);
LoadSprite(201, 25, 12, 12, SpFrame6);
LoadSprite(160,  0, 12, 24, SpFrame7);
LoadSprite(173,  0,  6, 24, SpFrame8);
LoadSprite(215,  0, 30, 24, SpFrame9);
LoadSprite(180, 38,  6, 30, SpFrameA);
LoadSprite(187, 38,  6, 30, SpFrameB);
LoadSprite(194, 38,  6, 30, SpFrameC);

LoadSprite(140, 70, 42, 14, addr SpButUp(0,0));
LoadSprite(186, 70, 42, 14, addr SpButUp(1,0));
LoadSprite(232, 70, 42, 14, addr SpButUp(2,0));
LoadSprite(232, 40, 42, 14, addr SpButUp(3,0));
LoadSprite(140, 85, 42, 14, addr SpButDn(0,0));
LoadSprite(186, 85, 42, 14, addr SpButDn(1,0));
LoadSprite(232, 85, 42, 14, addr SpButDn(2,0));
LoadSprite(232, 55, 42, 14, addr SpButDn(3,0));

I:= 0;
for Y:= 0, 8-1 do
    for X:= 0, 32-1 do
	begin
	LoadSprite(X*FontWidth, 100+Y*FontHeight, FontWidth, FontHeight,
		addr SpFont(I,0));
	I:= I + 1;
	end;

LoadPointer(120, 20, PtrCross);
LoadPointer(120, 40, PtrArrow);
LoadPointer(120, 60, PtrHand);
end;	\LoadSprites

\-------------------------------------------------------------------------------

proc	GetCmdLine(FN);
\Read any file name on the command line and add ".LIB" if no extension is given.
\Returns ".LAB" if there is no file name entered on the command line.
char	FN;		\address of string in which to return the file name
int	I, J, Ch, HaveExt;
char	Ext;
begin
Blit(PspSeg, $81, DataSeg, FN, $7F);	\copy command line into FN
I:= 0;
J:= 0;
HaveExt:= false;
loop	begin
	Ch:= FN(I);
	if Ch = CR then quit;
	if Ch = ^. then HaveExt:= true;
	I:= I + 1;
	if Ch # ^  then			\strip out space characters
		[FN(J):= Ch;  J:= J + 1];
	end;
if HaveExt then
	FN(J-1):= FN(J-1) ! $80		\terminate string
else	begin				\add extension if not given
	Ext:= ".LAB";
	for I:= 0, 3 do FN(I+J):= Ext(I);
	end;
end;	\GetCmdLine

\===============================================================================

begin	\Main
Trap(Debug);
CpuReg:= GetReg;
PspSeg:= CpuReg(11);
DataSeg:= CpuReg(12);
GetCmdLine(FileName);

if not OpenMouse then Exit("This program requires a mouse.");

SetVid($12);				\640x480 graphics
Screen(0):= Malloc(320/16*200);		\load sprites into temporary 64000-byte
LoadSprites;				\ hunk of memory, and set VGA color regs
Release(Screen(0));
Screen(0):= $A000;			\point to video memory

\NULL AND OR NOT BUT NAND NOR XOR LATCH FLIP NODE CLK TERM NAND2 NOR2 NOT2 LABEL
ObjWidth:= [26, 22, 22, 22, 22, 26, 26, 26, 22, 22,  4, 22,  4, 26, 26, 22, 18];
ObjHeight:=[17, 17, 17, 15,  7, 17, 17, 17, 17, 17,  5, 17,  7, 17, 17, 15, 10];

\The palette registers (which determine the colors) are the same for the two
\ pairs of bit planes. The other values (00) aren't used and don't matter.
CallInt($10, $1002, 0, 0,	\set palette registers and border color
	[$0500, $0F0A, $0005, $0000, $000A, $0000, $000F, $0000, $00],
	0, DataSeg, DataSeg);
ShowMask:= $3;
DrawMask:= $3;
SetVGA;

OpenPixBuf;

MovePen(0, 0);				\initialize (for safety)
ColorFg:= Black;   ColorBg:= White;
HaveGrid:= false;

DataSize:= 0;
Dirty:= false;
if FileName(0) # ^. then LoadData(FileName);

CallInt($33, 7, 0, 0, VWidth-2);	\set mouse pointer limit
CallInt($33, 4, 0, 50, 50);		\set mouse pointer near menus

DnD;	\(never returns)
end;	\Main
